#include "html2fo.h"
#include "structs.h"
#include "data.h"

/* *INDENT-OFF* */

/******************************************************
 *  complete internal data for converting             *
 *****************************************************/

tPropertyType DefPropertyTypes[] =
{
/*	{Name			, MapTo				, ValidInTags
													, pConvertPropertyType}
					  "@..."  See at ...									*/

	{"html-filepos"	, NULL				, NULL		, NULL},
	{"XML-NAMESPACE-fo"
					, "xmlns:fo"		, "HTML\0"	, NULL},
	{"XML-NAMESPACE-fox"
					, "xmlns:fox"		, "HTML\0"	, NULL},
	{"LANG"			, NULL				, "HTML\0"	, NULL},
	{"HTTP-EQUIV"	, NULL				, "META\0"	, &cptSetNull},
	{"CONTENT"		, "encoding"		, "META\0"	, NULL},
	{"NAME"			, NULL				, "META\0"	, &cptSetNull},
	{"FACE"			, NULL				, "FONT\0"	, NULL},
	{"XML-FONT-FACE"
					, "font-family"		, TG_FONTTAGS
													, NULL},
	{"SIZE"			, NULL				, "FONT\0"	, &cptFontSize},
	{"XML-FONT-SIZE"
					, "font-size"		, TG_FONTTAGS
													, NULL},
	{"XML-FONT-STYLE"
					, "font-style"		, TG_FONTTAGS
													, NULL},
	{"XML-FONT-WEIGHT"
					, "font-weight"		, TG_FONTTAGS
													, NULL},
	{"XML-FONT-TEXT-DECORATION"
					, "text-decoration"	, TG_FONTTAGS
													, NULL},
	{"XML-FONT-COLOR"
					, "color"			, TG_FONTTAGS
													, NULL},
	{"COLOR"		, "@STYLE-color"	, "FONT\0"	, NULL},
	{"XML-LINE-HEIGHT"
					, "line-height"		, TG_PARAGRAPH
										  "FONT\0"	, NULL},
	{"XML-SPACE-BEFORE"
					, "space-before.optimum"
										, TG_PARAGRAPH
													, NULL},
	{"XML-SPACE-AFTER"
					, "space-after.optimum"
										, TG_PARAGRAPH
													, NULL},
	{"LINK"			, NULL				, "BODY\0"	, &cptSetNull},
	{"VLINK"		, NULL				, "BODY\0"	, &cptSetNull},
	{"ALINK"		, NULL				, "BODY\0"	, &cptSetNull},
	{"TEXT"			, NULL				, "BODY\0"	, &cptSetNull},
	{"BORDER"		, NULL				, "IMG\0"	, &cptBorder},
	{"BORDER"		, NULL				, TG_BORDERTAGS
													, &cptTableBorder},
	{"BGCOLOR"		, "background-color"
										, "TABLE\0"
										  "TR\0"
										  "TD\0"
										  "BODY\0"	, NULL},
	{"BACKGROUND"	, "background-image"
										, "TD\0"	, &cptLink},
	{"BORDERCOLOR"	, NULL				, TG_BORDERTAGS
													, NULL},
	{"CELLSPACING"	, NULL				, "TABLE\0"
										  "TD\0"	, NULL},
	{"CELLPADDING"	, NULL				, "TABLE\0"
										  "TD\0"	, &cptPadding},
	{"TOPMARGIN"	, NULL				, "BODY\0"	, NULL},
	{"LEFTMARGIN"	, NULL				, "BODY\0"	, NULL},
	{"MARGINWIDTH"	, NULL				, "BODY\0"	, NULL},
	{"MARGINHEIGHT"	, NULL				, "BODY\0"	, NULL},
	{"XML-MARGIN-TOP"
					, "margin-top"		, TG_MARGINTAGS
													, NULL},
	{"XML-MARGIN-BOTTOM"
					, "margin-bottom"	, TG_MARGINTAGS
													, NULL},
	{"XML-MARGIN-LEFT"
					, "margin-left"		, TG_MARGINTAGS
													, NULL},
	{"XML-MARGIN-RIGHT"
					, "margin-right"	, TG_MARGINTAGS
													, NULL},
	{"WIDTH"		, NULL				, "TABLE\0"	, NULL},
	{"WIDTH"		, NULL				, "TD\0"	, NULL},
	{"WIDTH"		, "width"			, "IMG\0"	, &cptPoint},
	{"HEIGHT"		, "height"			, "TR\0"
										  "TD\0"
										  "IMG\0"	, &cptPoint},
	{"XML-NO-HEIGHT"
					, NULL				, "TD\0"	, NULL},
	{"ALIGN"		, "text-align"		, TG_BLOCK
										  "DIV\0"
										  "CENTER\0"
													, &cptAlign},
	{"ALIGN"		, NULL				, "TD\0"
										  "TR\0"	, &cptAlign},
	{"VALIGN"		, NULL				, "P\0"
										  "TD\0"
										  "TR\0"	, NULL},
	{"XML-WORDWRAP"	, "wrap-option"		, "XML-PAGE-SEQUENCE\0"
										  TG_BLOCK
										  "NOBR\0"	, NULL},
	{"XML-HYPHENATE"
					, "hyphenate"		, "XML-PAGE-SEQUENCE\0"
										  TG_BLOCK	, NULL},
	{"XML-KEEP-TOGETHER"
					, "keep-together"	, TG_BLOCK	, NULL},
	{"XML-WHITE-SPACE-COLLAPSE"
					, "white-space-collapse"
										, "P\0"
										  "PRE\0"	, NULL},
	{"XML-ID"		, "id"				, TG_BLOCK
										  TG_INLINE
													, NULL},
/*	{"XML-BLOCK-ALIGN"
					, "text-align"		, "P\0"
										  "BR\0"	, NULL},
*/	{"COLSPAN"		, "number-columns-spanned"
										, "TD\0"
                                          "TH\0"	, NULL},
	{"ROWSPAN"		, "number-rows-spanned"
										, "TD\0"	, NULL},
	{"REMOVED-ROWSPAN"		, NULL/*"number-rows-spanned"*/
										, "TD\0"	, NULL},
	{"XML-TABLE-COLUMN-WIDTH"
					, "column-width"	, "XML-TABLE-COLUMN\0"
													, NULL},
	{"XML-TABLE-LAYOUT"
					, "table-layout"	, "TABLE\0"	, NULL},
	{"WIDTH"		, "@XML-TABLE-COLUMN-WIDTH"
										, "XML-TABLE-COLUMN\0"
													, NULL},
/*	{"XML-CELL-WIDTH"
					, "width"			, "TD\0"	, NULL},
*/
	{"XML-BORDER-WIDTH"
					, "border-width"	, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-COLOR"
					, "border-color"	, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-STYLE"
					, "border-style"	, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-LEFT-WIDTH"
					, "border-left-width"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-RIGHT-WIDTH"
					, "border-right-width"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-TOP-WIDTH"
					, "border-top-width"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-BOTTOM-WIDTH"
					, "border-bottom-width"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-LEFT-COLOR"
					, "border-left-color"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-RIGHT-COLOR"
					, "border-right-color"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-TOP-COLOR"
					, "border-top-color"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-BOTTOM-COLOR"
					, "border-bottom-color"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-LEFT-STYLE"
					, "border-left-style"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-RIGHT-STYLE"
					, "border-right-style"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-TOP-STYLE"
					, "border-top-style"
										, TG_BORDERTAGS
													, NULL},
	{"XML-BORDER-BOTTOM-STYLE"
					, "border-bottom-style"
										, TG_BORDERTAGS
													, NULL},
	{"--"			, "--"				, "!--\0"	, &cptSetNull},
	{"--Creator"	, "Creator"			, "!--\0"	, NULL},
	{"--Version"	, "Version"			, "!--\0"	, NULL},
	{"XML-?-VERSION"
					, "version"			, "XML-?\0"	, NULL},
	{"XML-?-ENCODING"
					, "encoding"		, "XML-?\0"	, NULL},
	{"XML-?-?"		, "?"				, "XML-?\0"	, NULL},
	{"XML-PAGE-HEIGHT"
					, "page-height"		, "XML-SIMPLE-PAGE-MASTER\0"
													, NULL},
	{"XML-PAGE-WIDTH"
					, "page-width"		, "XML-SIMPLE-PAGE-MASTER\0"
													, NULL},
	{"XML-MARGIN-TOP"
					, "margin-top"		, "XML-SIMPLE-PAGE-MASTER\0"
										  "XML-REGION-BODY\0"
													, NULL},
	{"XML-MARGIN-BOTTOM"
					, "margin-bottom"	, "XML-SIMPLE-PAGE-MASTER\0"
										  "XML-REGION-BODY\0"
													, NULL},
	{"XML-MARGIN-LEFT"
					, "margin-left"		, "XML-SIMPLE-PAGE-MASTER\0"
										  "XML-REGION-BODY\0"
													, NULL},
	{"XML-MARGIN-RIGHT"
					, "margin-right"	, "XML-SIMPLE-PAGE-MASTER\0"
										  "XML-REGION-BODY\0"
													, NULL},
	{"XML-REGION-EXTENT"
					, "extent"			, "XML-REGION-BEFORE\0"
										  "XML-REGION-AFTER\0"
													, NULL},
	{"XML-PAGE-SEQUENCE-LANGUAGE"
					, "language"		, "XML-PAGE-SEQUENCE\0"
													, NULL},
	{"XML-SIMPLE-PAGE-MASTER-NAME"
					, "master-name"
										, "XML-SIMPLE-PAGE-MASTER\0"
													, NULL},
	{"XML-PAGE-SEQUENCE-MASTER-REFERENCE"
					, "master-reference"
										, "XML-PAGE-SEQUENCE\0"
													, NULL},
	{"BODY-XML-FLOW-NAME"
					, "flow-name"		, "BODY\0"
										  "XML-STATIC-CONTENT\0"
													, NULL},
	{"NAME"			, "id"				, "A\0"		, &cptLinkName},
	{"HREF"			, "external-destination"
										, "A\0"		, &cptLink},
	{"XML-INTERNAL-DESTINATION"
					, "internal-destination"
										, "A\0"		, &cptLink},
	{"TARGET"		, NULL				, "A\0"		, NULL},
	{"SRC"			, "src"				, "IMG\0"	, NULL},
	{"NAME"			, NULL				, "IMG\0"	, NULL},
	{"ALT"			, NULL				, "IMG\0"	, NULL},
	{"ALIGN"		, NULL				, "IMG\0"	, NULL},
	{"WIDTH"		, NULL				, "IMG\0"	, NULL},
	{"HEIGHT"		, NULL				, "IMG\0"	, NULL},
	{"BORDER"		, NULL				, "IMG\0"	, NULL},
	{"HSPACE"		, NULL				, "IMG\0"	, NULL},
	{"VSPACE"		, NULL				, "IMG\0"	, NULL},
	{"ONMOUSEOVER"	, NULL				, "A\0"		, NULL},
	{"ONMOUSEOUT"	, NULL				, "A\0"		, NULL},
	{"ONLOAD"		, NULL				, "IMG\0"	, NULL},
	{"EEID"			, NULL				, NULL		, NULL},
	{"LIST-SEPARATOR"
					, "\0"				, NULL		, NULL},
	{"CLASS"		, NULL				, NULL		, NULL},
	{"TITLE"		, NULL				, "HR\0"
										  "ABBR\0"	, NULL},
	{"LANG"			, NULL				, "ABBR\0"	, NULL},
	{"REL"			, NULL				, "LINK\0"	, NULL},
	{"TYPE"			, NULL				, "LINK\0"
										  "STYLE\0"	, NULL},
	{"HREF"			, NULL				, "LINK\0"	, NULL},

	{"XML-leader-pattern"
					, "leader-pattern"	, "HR\0"	, NULL},
	{"XML-rule-thickness"
					, "rule-thickness"	, "HR\0"	, NULL},
	{"XML-space-before.optimum"
					, "space-before.optimum"
										, "HR\0"	, NULL},
	{"XML-space-after.optimum"
					, "space-after.optimum"
										, "HR\0"	, NULL},
	{"XML-start-indent"
					, "start-indent"	, "XML-LIST-ITEM-BODY\0"
										  "HR\0"	, NULL},
	{"XML-end-indent"
					, "end-indent"		, "HR\0"	, NULL},
	{"XML-provisional-distance-between-starts"
					, "provisional-distance-between-starts"
										, "UL\0"
                                          "OL\0"	, NULL},
	{"XML-provisional-label-separation"
					, "provisional-label-separation"
										, "UL\0"
                                          "OL\0"	, NULL},
/* STYLE-Properties */
	{"STYLE"		, NULL				, TG_FONTTAGS
										  "A\0"
										  "DIV\0"
										  "TABLE\0"	, &cptParseProperties},
	{"STYLE-color"
					, NULL				, TG_FONTTAGS
													, NULL},
	{"STYLE-font-size"
					, NULL				, TG_FONTTAGS
													, NULL},
	{"STYLE-font-family"
					, NULL				, TG_FONTTAGS
													, NULL},
	{"STYLE-font-style"
					, NULL				, TG_FONTTAGS
													, NULL},
	{"STYLE-align"
					, NULL				, TG_FONTTAGS
													, NULL},
	{"STYLE-line-height"
					, "@XML-LINE-HEIGHT"
										, TG_BLOCK_NO_PG
										  TG_PARAGRAPH
													, NULL},
	{"STYLE-margin-top"
					, "margin-top"		, "P\0"		, NULL},
	{"STYLE-margin-bottom"
					, "margin-bottom"	, "P\0"		, NULL},
	{"STYLE-margin-left"
					, "margin-left"		, "P\0"		, NULL},
	{"STYLE-margin-right"
					, "margin-right"	, "P\0"		, NULL},
	{"STYLE-page-break-before"
					, "break-before"
										, TG_BLOCK
										  "TABLE\0"	, &cptGenericBreak},
	{"STYLE-page-break-inside"
					, "break-inside"
										, TG_BLOCK
										  "TABLE\0"	, &cptGenericBreak},
	{"STYLE-page-break-after"
					, "break-after"
										, TG_BLOCK
										  "TABLE\0"	, &cptGenericBreak},
	{"STYLE-text-decoration"
					, "text-decoration"
										, "P\0"
										  "A\0"		, NULL},
	{"STYLE-color"
					, "color"
										, "P\0"
										  "A\0"		, NULL},

/* Configuration Properties */
	{"XML-CONFIG"	, NULL				, NULL		, &cptParseProperties},
	{"XML-CONFIG-convert-row-spans"
					, NULL				, "HTML\0"
										  "BODY\0"
										  "TABLE\0"	, NULL},
	{"XML-CONFIG-convert-table-borders"
					, NULL				, "HTML\0"
										  "BODY\0"
										  "TABLE\0"	, NULL},
	{"XML-CONFIG-convert-links"
					, NULL				, NULL		, NULL},
	{"XML-CONFIG-max-content-width"
					, NULL				, "BODY\0"
										  "TD\0"	, NULL},
	{NULL}
};

tPTag ROOT = NULL;

tTagType DefTagTypes[] =
{
/*	{Name		, MapTo				, EndTag
				  "\0"    Ignore			, Cascade
				  "@..."  Processing like ...		, CloseXMLTag
															, ConvertProperties
																	, SetupIndent
																			, CloseOtherTags
																					, ValidSubTags	, pProcessTag		, ProcessBefore}*/

	{"XML-?"	, "?xml"			, No	, No	, No	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"HTML"		, "fo:root"			, Yes	, No	, Yes	, Yes	, Yes	, NULL	, "HEAD\0"
																					  "XML-LAYOUT-MASTER-SET\0"
																					  "XML-PAGE-SEQUENCE\0"
																					  "BODY\0"
																					  "XML-STATIC-CONTENT\0"
																					  "!--\0"		, &ProcessTagHTML},
	{"XML-LAYOUT-MASTER-SET"
				, "fo:layout-master-set"
									, No	, No	, Yes	, Yes	, Yes	, NULL	, "XML-SIMPLE-PAGE-MASTER\0"
																									, NULL},
	{"XML-SIMPLE-PAGE-MASTER"
				, "fo:simple-page-master"
									, No	, No	, Yes	, Yes	, Yes	, NULL	, "XML-REGION-BEFORE\0"
																					  "XML-REGION-BODY\0"
																					  "XML-REGION-AFTER\0"
																									, NULL},
	{"XML-REGION-BEFORE"
				, "fo:region-before"
									, No	, No	, MayBe	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"XML-REGION-BODY"
				, "fo:region-body"	, No	, No	, MayBe	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"XML-REGION-AFTER"
				, "fo:region-after"	, No	, No	, MayBe	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"HEAD"		, NULL				, Yes	, No	, Yes	, Yes	, Yes	, NULL	, "META\0"
																					  "TITLE\0"
																					  "Footer\0"
																					  "STYLE\0"
																					  "SCRIPT\0"
																					  "LINK\0"
																					  "!--\0"		, NULL},
	{"META"		, NULL				, No	, No	, Yes	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"TITLE"	, NULL				, Yes	, No	, Yes	, Yes	, Yes	, NULL	, "\0"			, &ProcessTagTitle},
	{"Footer"	, NULL				, Yes	, No	, Yes	, Yes	, Yes	, NULL	, "\0"			, &ProcessTagFooter},
	{"LINK"		, NULL				, No	, No	, Yes	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"BODY"		, "fo:flow"			, Yes	, No	, Yes	, Yes	, Yes	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE
																					  "!--\0"		, &ProcessTagBody	, &ProcessTagBodyBefore},
	{"XML-STATIC-CONTENT"
				, "fo:static-content"
									, No	, No	, Yes	, Yes	, Yes	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"XML-PAGE-SEQUENCE"
				, "fo:page-sequence"
									, Yes	, No	, Yes	, Yes	, Yes	, NULL	, "BODY\0"		, NULL},
	{"P"		, "fo:block"		, MayBe	, No	, Yes	, Yes	, No	, "BR\0"
																					, TG_BLOCK
																					  TG_FONT
																					  TG_INLINE		, NULL},
	{"NOBR"		, "fo:inline"
									, No	, No	, Yes	, Yes	, No	, NULL	, NULL			, NULL},
	{"BR"		, "fo:inline white-space-collapse=\"false\">\n</fo:inline"
									, No	, No	, No	, Yes	, No	, NULL	, NULL			, NULL},
	{"XML-INLINE"
				, "fo:inline"		, Yes	, No	, Yes	, Yes	, No	, NULL	, "\0"			, NULL},
	{"FONT"		, NULL				, Yes	, Yes	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE
																					  "PRE\0"		, NULL},
	{"TABLE"	, "fo:table"		, Yes	, MayBe	, Yes	, Yes	, Yes	, TG_INLINE
																					, "TR\0"
																					  "XML-TABLE-COLUMN\0"
																					  "XML-TABLE-BODY\0"
																					  "XML-TABLE-FOOTER\0"
																					  "COLGROUP\0"
																					  "CAPTION\0"	, &ProcessTagTable	, &ProcessTagTableBefore},
	{"CAPTION"	, "\0"				, Yes	, No	, No	, No	, Yes	, NULL	, NULL			, NULL},
	{"COL"		, "@XML-TABLE-COLUMN"
									, No	, No	, No	, No	, Yes	, NULL	, NULL			, NULL},
	{"XML-TABLE-COLUMN"
				, "fo:table-column"	, No	, No	, MayBe	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"XML-TABLE-FOOTER"
				, "fo:table-footer"	, Yes	, MayBe	, Yes	, Yes	, Yes	, NULL	, "TR\0"		, NULL},
	{"TBODY"	, "\0"
									, No	, No	, No	, No	, Yes	, NULL	, NULL			, NULL},
	{"XML-TABLE-BODY"
				, "fo:table-body"	, Yes	, MayBe	, Yes	, Yes	, Yes	, NULL	, "TR\0"		, NULL},
	{"TH"		, "@TD"				, Yes	, MayBe	, Yes	, Yes	, Yes	, NULL	, "FONT\0"		, NULL},
	{"TR"		, "fo:table-row"	, Yes	, MayBe	, Yes	, Yes	, Yes	, NULL	, "TD\0"
																					  "TH\0"
																					  "XML-EMPTY-CELL\0"
																									, NULL},
	{"TD"		, "fo:table-cell"	, Yes	, MayBe	, Yes	, Yes	, Yes	, NULL	, "BR\0"
																					  TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"XML-EMPTY-CELL"
				, "fo:table-cell"	, Yes	, MayBe	, Yes	, Yes	, Yes	, NULL	, "\0"			, NULL},
	{"XML-LINK-REFERENCE"
				, "fo:inline"		, Yes	, No	, Yes	, Yes	, No	, NULL	, NULL			, NULL},
	{"A"		, "fo:basic-link"	, Yes	, No	, Yes	, Yes	, No	, "A\0"	, NULL			, &ProcessTagA},
	{"U"		, NULL				, Yes	, Yes	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"B"		, NULL				, Yes	, Yes	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"I"		, NULL				, Yes	, Yes	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"DIR"		, NULL				, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, NULL},
	{"DIV"		, "fo:block"		, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_BLOCK
																					  TG_FONT
																					  TG_INLINE		, NULL},
	{"!--"		, "!--"				, No	, No	, No	, No	, No	, NULL	, "\0"			, NULL},
	{"!DOCTYPE"	, NULL				, No	, No	, No	, No	, Yes	, NULL	, "\0"			, NULL},
	{"COLGROUP"	, "\0"				, Yes	, MayBe	, No	, Yes	, Yes	, NULL	, "COL\0"		, NULL},
	{"HR"		, "fo:leader"		, No	, No	, Yes	, Yes	, No	, NULL	, "\0"			, NULL},
	{"H1"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, &ProcessTagHeader},
	{"H2"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, &ProcessTagHeader},
	{"H3"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, &ProcessTagHeader},
	{"H4"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, &ProcessTagHeader},
	{"H5"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, &ProcessTagHeader},
	{"H6"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, &ProcessTagHeader},
	{"PRE"		, "fo:block"		, Yes	, No	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_INLINE		, NULL},
	{"OL"		, "fo:list-block"	, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, "LI\0"		, NULL},
	{"UL"		, "fo:list-block"	, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, "LI\0"		, NULL},
	{"LI"		, "fo:list-item"	, MayBe	, MayBe	, Yes	, Yes	, No	, "LI\0"
																					, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE
																					  "OL\0"
																					  "UL\0"
																					  "DL\0"
																					  "XML-LIST-ITEM-LABEL\0"
																					  "XML-LIST-ITEM-BODY\0"
																									, &ProcessTagListItem},
	{"XML-LIST-ITEM-LABEL"
			, "fo:list-item-label"	, No	, MayBe	, Yes	, Yes	, Yes	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"XML-LIST-ITEM-BODY"
			, "fo:list-item-body"	, No	, MayBe	, Yes	, Yes	, Yes	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"IMG"	, "fo:external-graphic"
									, No	, No	, MayBe	, Yes	, No	, NULL	, NULL			, NULL},
	{"CENTER"
			, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"STRONG"
			, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"SMALL"
			, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"SCRIPT"
			, NULL					, Yes	, No	, Yes	, No	, No	, NULL	, NULL			, NULL},
	{"STYLE"
			, NULL					, Yes	, No	, Yes	, Yes	, No	, NULL	, NULL			, &ProcessTagNullSubInfo},
	{"CODE"	, NULL					, Yes	, No	, Yes	, Yes	, No	, NULL	, NULL			, NULL},
	{"DL"	, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, "DT\0"
																					  "DD\0"
																					  "LI\0"
																					  "P\0"
																					  "BLOCKQUOTE\0"
																									, NULL},
	{"DD"	, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"DT"	, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{"ABBR"	, NULL					, Yes	, No	, Yes	, Yes	, No	, NULL	, "\0"			, NULL},
	{"SUP"	, NULL					, Yes	, No	, Yes	, Yes	, No	, NULL	, "\0"			, NULL},
	{"SUB"	, NULL					, Yes	, No	, Yes	, Yes	, No	, NULL	, "VAR\0"		, NULL},
	{"VAR"	, NULL					, Yes	, No	, Yes	, Yes	, No	, NULL	, "\0"			, NULL},
	{"BLOCKQUOTE"
			, NULL					, Yes	, MayBe	, Yes	, Yes	, No	, NULL	, TG_FONT
																					  TG_BLOCK
																					  TG_INLINE		, NULL},
	{NULL}
};

tTagGroup DefTagGroups[] =
{
	{"%BLOCK;"		, TG_BLOCK_NO_PG
					  TG_PARAGRAPH
					  TG_BLOCK_FB},

	{"%BLOCK-NO-PG;"
					, "DIV\0"
					  "!--\0"
					  "PRE\0"
					  "IMG\0"
					  "BLOCKQUOTE\0"},

	{"%PARAGRAPH;"	, "P\0"
					  "H1\0"
					  "H2\0"
					  "H3\0"
					  "H4\0"
					  "H5\0"
					  "H6\0"},

	{"%BLOCK-FB;"	, "TABLE\0"
					  "OL\0"
					  "UL\0"
					  "DL\0"},

	{"%INLINE;"		, "XML-INLINE\0"
					  "A\0"
					  "BR\0"
					  "NOBR\0"
					  "IMG\0"
					  "ABBR\0"
					  "SUP\0"
					  "SUB\0"
					  "VAR\0"
					  "XML-LINK-REFERENCE\0"
					  "HR\0"
					  "!--\0"},

	{"%FONT;"		, "FONT\0"
					  "B\0"
					  "U\0"
					  "I\0"
					  "SMALL\0"
					  "STRONG\0"
					  "CENTER\0"
					  "RIGHT\0"
					  "CODE\0"
					  "BLOCKQUOTE\0"},

	{"%MARGINTAGS;"	, "TABLE\0"
					  "TD\0"},

	{"%BORDERTAGS;"	, "TABLE\0"
					  "TD\0"
					  "TH\0"
					  "TR\0"
					  "XML-TABLE-BODY\0"
					  "XML-TABLE-COLUMN\0"
					  "IMG\0"},

	{"%FONTTAGS;"	, "XML-INLINE\0"
					  "FONT\0"
					  "CODE\0"
					  "PRE\0"
					  "A\0"
					  "TD\0"
					  TG_PARAGRAPH},
	{NULL}
};

tPreDefProperty DefPreDefProperties[] =
{
/*	{Tag			, Property									, Value}*/
	{"CODE"			, "XML-FONT-FACE"							, "Courier"},
	{"HR"			, "XML-leader-pattern"						, "rule"},
	{"HR"			, "XML-rule-thickness"						, "2.0pt"},
	{"HR"			, "XML-space-before.optimum"				, "0pt"},
	{"HR"			, "XML-space-after.optimum"					, "0pt"},
	{"HR"			, "XML-start-indent"						, "0cm"},
	{"HR"			, "XML-end-indent"							, "0cm"},
	{"PRE"			, "XML-white-space-collapse"				, "false"},
	{"PRE"			, "XML-FONT-FACE"							, "Courier"},
	{"OL"			, "XML-provisional-distance-between-starts"	, "14pt"},
	{"OL"			, "XML-provisional-label-separation"		, "3pt"},
	{"UL"			, "XML-provisional-distance-between-starts"	, "14pt"},
	{"UL"			, "XML-provisional-label-separation"		, "3pt"},
	{"TABLE"		, "XML-TABLE-LAYOUT"						, "fixed"},
	{"TABLE"		, "ALIGN"									, "left"},
	{"NOBR"			, "XML-WORDWRAP"							, "no-wrap"},
	{"CENTER"		, "ALIGN"									, "center"},
	{NULL}
};

tSpecialChar DefSpecialChar[] =
{
	{"&nbsp;"	," "	,"<fo:inline white-space-collapse=\"false\">"	,"</fo:inline>"},
	{"&quot;"	,NULL},
	{"&auml;"	,""},
	{"&ouml;"	,""},
	{"&uuml;"	,""},
	{"&Auml;"	,""},
	{"&Ouml;"	,""},
	{"&Uuml;"	,""},
	{"&szlig;"	,""},
	{"&gt;"		,NULL},
	{"&lt;"		,NULL},
	{"&sect;"	,NULL},
	{"&amp;"	,NULL},
	{"&copy;"	,""},
	/* all other '&'s should be masked */
	{"&"		,"&#38;"},
	{NULL}
};

tPChar LangCodes[][2] =
{
	{
		"en\0"
		"de\0"	, "ISO-8859-1"
	},
	{NULL}
};

/* *INDENT-ON* */

