#include <errno.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "data.h"
#include "config.h"

/******************************************************
*  Globale Variablen und #define's                    *
******************************************************/

#if defined(WIN32) || defined(WINDOWS) || defined(DOS)

/* needed for compatiblity */

#define strcasecmp stricmp
#define NullDevice "nul"

#else

#define NullDevice "/dev/null"

#endif


#define DepthIndent 4
extern int memblocks;
extern FILE *mylog, *myout;
extern char WarnProperties;
extern char myinputfile[100];

#define false 0
#define true !false

#define IsSpace(Str) (strcspn(Str," \t\n\r")==0)
#define malloctype(t) (t *)malloc(sizeof(t))

#define NullString(str) ((&*str)?str:"")

extern char debmodeorig;

/******************************************************
*  Struktur- und Typdefinitionen                      *
******************************************************/

/*

Referenzen der Tags:

  Parent<---------.
                  |
          ,---,---
          | S |<--- Parent-referenzen von SubTags
          `---
       ,- ||
       |   |`--->SubTags
      .|,  `---->LastSubTag
     Next

,---.
| 1 |<---------.
`---          |
 LS    ,---,P-|
 |`--->| 2 |   |
 |     `---   |
 |  ,N        |
 |  |          |
 |  |  ,---,P--
 |  `->| 3 |
 |`--->`---
 |  ,-
 ...

S SubTags
L LastSubTag
P Parent
N Next

InsertAsSubTag
,---.
| 1 |<---------. <- Referenz
`---          |
 ||    ,---,---
 |`--->| i |<---------.
 `---->`---          |
        ||    ,---,--|
        |`--->| 2 |   |
        |     `---   |
        |  ,-        |
        |  |  ,---,---
        |  `->| 3 |
        `---->`---

InsertIncTag
,---.
| 1 |<---------.
`---          |
 ||    ,---,--|
 |`--->| 2 |   | <- Referenz
 |     `---   |
 |  ,-        |
 |  |          |
 |  |  ,---,---
 |  `->| i |<---------.
 |     `---          |
 |  ,- ||    ,---,---
 ...    |`--->| 3 |
        `---->`---

AddSubTagFirst
,---.
| 1 |<---------. <- Referenz
`---          |
 ||    ,---,--|
 |`--->| i |   |
 |     `---   |
 |  ,-        |
 |  |  ,---,--|
 |  `->| 2 |   |
 |     `---   |
 |  ,-        |
 |  |  ,---,---
 |  `->| 3 |
 `---->`---

AddTagAfter
,---.
| 1 |<---------.
`---          |
 ||    ,---,--|
 |`--->| 2 |   | <- Referenz
 |     `---   |
 |  ,-        |
 |  |  ,---,--|
 |  `->| i |   |
 |     `---   |
 |  ,-        |
 |  |  ,---,---
 |  `->| 3 |
 `---->`---

AddTagAfterIncRest
,---.
| 1 |<---------.
`---          |
 ||    ,---,--|
 |`--->| 2 |   | <- Referenz
 |     `---   |
 |  ,-        |
 |  |  ,---,--
 |  `->| i |<--------.
 `---->`---         |
        ||           |
        ||    ,---,--
        |`--->| 3 |
        `---->`---

AddSubTagLast
,---.
| 1 |<---------. <- Referenz
`---          |
 ||    ,---,--|
 |`--->| 2 |   |
 |     `---   |
 |  ,-        |
 |  |  ,---,--|
 |  `->| 3 |   |
 |     `---   |
 |  ,-        |
 |  |  ,---,---
 |  `->| i |
 `---->`---
*/



/* ***** Functions ***** */


long a2l(tPChar Value, long StdValue);
double a2f(tPChar Value, double StdValue);
double ap2f(tPChar Value, double StdValue, double percentbase);
int strmulticmp(tPChar find, tPChar multi);
int isspacestr(char *s);

tPPropertyType SearchPropertyType(tPChar PropertyName, tPTagType pTagType);
tPProperty SearchProperty(tPTag pTag, tPChar Name);
tPTagType SearchTagType(tPTag pParentTag, tPChar TagName);
void ParseProperty(tPProperty * ppProp, tPChar Name, tPChar Value,
				   tPTagType pTagType);

char InitROOT();

tPTag InsertAsSubTag(tPTag pTag, tPChar TagName);
tPTag InsertToSubTag(tPTag pTag, tPChar TagName);
tPTag AddSubTagFirstIncludeText(tPTag pTag, tPChar TagName);
tPTag AddSubTagFirst(tPTag pTag, tPChar TagName);
tPTag AddTagBefore(tPTag pTag, tPChar TagName);
tPTag AddTagAfter(tPTag pTag, tPChar TagName);
tPTag AddTagAfterIncRest(tPTag pTag, tPChar TagName);
tPTag AddSubTagLast(tPTag pTag, tPChar TagName);

tPProperty RemoveProperty(tPTag pTag, tPProperty pProp);
tPProperty AddProperty(tPTag pTag, tPChar PropertyName);
tPProperty AddPropertyValue(tPTag pTag, tPChar PropertyName,
							tPChar PropertyValue);
void AddTreePropertyValue(tPTag pTag, tPChar PropertyName,
						  tPChar PropertyValue, char ProcessSubAfterHit,
						  char ProcessNext);
tPProperty AddPropertySeparator(tPTag pTag);

void FreeTag(tPPTag ppTag);

char IsTrivialTree(int depth, tPTag pt);
tPTag GetPrevTag(tPTag pt);
tPChar getIndent(tPTag pTag);
void printTagFilePos(FILE * f, tPTag pt);
tPProperty GetConfig(tPTag pt, tPChar config);

