#include "html2fo.h"

/******************************************************
 *  globals                                           *
 *****************************************************/

char debmodeorig;
char debon;
int memblocks;
char *FunctionName = "";
void *memblock;
FILE *mylog, *myout;
char WarnProperties = true;
char myinputfile[100];
tPPChar fbuf;
long fbuflen = 0;
tPChar fbufbegin = NULL;
tPTag TagStack[50] = { NULL };
int TagStackPos = 0;
long tagstartpos = 0;

/******************************************************
 *  Helpers for accessing data                        *
 *****************************************************/

tPProperty SearchProperty(tPTag pTag, tPChar Name)
{
	tPProperty pos;

	if (pTag == NULL)
		return NULL;
	pos = pTag->Properties;
	while (pos != NULL)
	{
		if (strcasecmp(pos->Type->Name, Name) == 0)
		{
			break;
		}
		pos = pos->Next;
	}
	return pos;
}

tPTagType SearchTagType(tPTag pParentTag, tPChar TagName)
{
	int ttpos = 0;
	tPTagType ptt = NULL;

	while (DefTagTypes[ttpos].Name != NULL)
	{
		if (strcasecmp(DefTagTypes[ttpos].Name, TagName) == 0)
		{
			if ((pParentTag == NULL) ||
				(pParentTag->Type->ValidSubTags == NULL) ||
				(strmulticmp(TagName, pParentTag->Type->ValidSubTags) == 0))
			{
				ptt = &DefTagTypes[ttpos];
				break;
			}
		}
		ttpos++;
	}
	return ptt;
}

tPPropertyType SearchPropertyType(tPChar PropertyName, tPTagType pTagType)
{
	int tppos = 0;
	tPPropertyType ppt = NULL;

	while (DefPropertyTypes[tppos].Name != NULL)
	{
		if (strcasecmp(PropertyName, DefPropertyTypes[tppos].Name) == 0)
		{
			if ((DefPropertyTypes[tppos].ValidInTags == NULL) ||
				(strmulticmp
				 (pTagType->Name, DefPropertyTypes[tppos].ValidInTags) == 0))
			{
				ppt = &DefPropertyTypes[tppos];
				break;
			}
		}
		tppos++;
	}

	return ppt;
}

tPProperty GetConfig(tPTag pt, tPChar config)
{
	tPProperty pp = NULL;

	if (pt != NULL)
	{
		pp = SearchProperty(pt, config);
		if (pp == NULL)
		{
			if (pt->Parent != NULL)
			{
				pp = GetConfig(pt->Parent, config);
			}
		}
	}

	return pp;
}

tPTag InsertAsSubTag(tPTag pTag, tPChar TagName)
{
	tPTag pt = pTag, ptins = NULL, ptsub;
	tPTagType ptt;

	ptt = SearchTagType(pTag, TagName);

	if (ptt != NULL)
	{
		ptins = malloctype(tTag);
		memset(ptins, 0, sizeof(tTag));
		ptins->Type = ptt;
		ptins->Parent = pt;
		ptins->SubTags = pt->SubTags;
		ptins->LastSubTag = pt->LastSubTag;
		ptins->TextIn = pt->TextIn;
		pt->TextIn = NULL;
		pt->SubTags = ptins;
		pt->LastSubTag = ptins;
		ptsub = ptins->SubTags;

		while (ptsub != NULL)
		{
			ptsub->Parent = ptins;
			ptsub = ptsub->Next;
		}
	}

	return ptins;
}

tPTag InsertToSubTag(tPTag pTag, tPChar TagName)
{
	tPTag ptp = NULL, ptins = NULL;
	tPTagType ptt;
	char FirstElement = false;

	if ((pTag != NULL) && (pTag->Parent != NULL))
		ptp = pTag->Parent->SubTags;

	if (ptp == pTag)
	{
		FirstElement = true;
	}
	else
		while ((ptp != NULL) && (ptp->Next != pTag))
			ptp = ptp->Next;

	if (ptp != NULL)
	{
		ptt = SearchTagType(pTag->Parent, TagName);

		if (ptt != NULL)
		{
			ptins = malloctype(tTag);
			memset(ptins, 0, sizeof(tTag));
			ptins->Type = ptt;

			if (pTag->Parent->LastSubTag == pTag)
			{
				pTag->Parent->LastSubTag = ptins;
			}
			ptins->Parent = pTag->Parent;
			ptins->SubTags = pTag;
			ptins->LastSubTag = pTag;
			ptins->Next = pTag->Next;
			ptins->TextAfter = pTag->TextAfter;
			pTag->TextAfter = NULL;
			pTag->Next = NULL;
			pTag->Parent = ptins;
			if (FirstElement)
			{
				ptins->Parent->SubTags = ptins;
			}
			else
			{
				ptp->Next = ptins;
			}

		}
	}
	return ptins;
}

tPTag AddSubTagFirstIncludeText(tPTag pTag, tPChar TagName)
{
	tPTag pt = pTag, ptins = NULL;
	tPTagType ptt;

	ptt = SearchTagType(pTag, TagName);

	if (ptt != NULL)
	{
		ptins = malloctype(tTag);
		memset(ptins, 0, sizeof(tTag));
		ptins->Type = ptt;
		ptins->Parent = pt;
		ptins->Next = pt->SubTags;
		ptins->TextIn = pt->TextIn;
		pt->TextIn = NULL;
		pt->SubTags = ptins;

		if (pt->LastSubTag == NULL)
		{
			pt->LastSubTag = ptins;
		}
	}

	return ptins;
}

tPTag AddSubTagFirst(tPTag pTag, tPChar TagName)
{
	tPTag pt = pTag, ptins = NULL;
	tPTagType ptt;

	ptt = SearchTagType(pTag, TagName);

	if (ptt != NULL)
	{
		ptins = malloctype(tTag);
		memset(ptins, 0, sizeof(tTag));
		ptins->Type = ptt;
		ptins->Parent = pt;
		ptins->Next = pt->SubTags;
		pt->SubTags = ptins;

		if (pt->LastSubTag == NULL)
		{
			pt->LastSubTag = ptins;
		}
	}

	return ptins;
}

tPTag AddSubTagLast(tPTag pTag, tPChar TagName)
{
	tPTag pt = pTag, ptins = NULL;
	tPTagType ptt;

	ptt = SearchTagType(pTag, TagName);

	if (ptt != NULL)
	{
		ptins = malloctype(tTag);
		memset(ptins, 0, sizeof(tTag));
		ptins->Type = ptt;
		ptins->Parent = pt;
		if (pt->LastSubTag != NULL)
		{
			pt->LastSubTag->Next = ptins;
		}
		else
		{
			pt->SubTags = ptins;
		}

		pt->LastSubTag = ptins;
	}

	return ptins;
}

tPTag AddTagBefore(tPTag pTag, tPChar TagName)
{
	tPTag pt = GetPrevTag(pTag);

	if (pt == NULL)
	{
		return AddSubTagFirst(pTag->Parent, TagName);
	}
	else
	{
		return AddTagAfter(pt, TagName);
	}
}

tPTag AddTagAfter(tPTag pTag, tPChar TagName)
{
	tPTag pt = pTag, ptins = NULL;
	tPTagType ptt;

	ptt = SearchTagType(pTag->Parent, TagName);

	if (ptt != NULL)
	{
		ptins = malloctype(tTag);
		memset(ptins, 0, sizeof(tTag));
		ptins->Type = ptt;
		ptins->Parent = pt->Parent;
		ptins->Next = pt->Next;
		pt->Next = ptins;

		if (ptins->Next == NULL)
		{
			ptins->Parent->LastSubTag = ptins;
		}
	}

	return ptins;
}

tPTag AddTagAfterIncRest(tPTag pTag, tPChar TagName)
{
	tPTag pt = pTag, ptins = NULL, ptsub;
	tPTagType ptt;

	ptt = SearchTagType(pTag->Parent, TagName);

	if (ptt != NULL)
	{
		ptins = malloctype(tTag);
		memset(ptins, 0, sizeof(tTag));
		ptins->Type = ptt;
		ptins->Parent = pt->Parent;
		ptins->SubTags = pt->Next;
		ptsub = pt->Next;
		pt->Next = ptins;
		pt->Parent->LastSubTag = ptins;

		while (ptsub != NULL)
		{
			ptsub->Parent = ptins;
			ptins->LastSubTag = ptsub;
			ptsub = ptsub->Next;
		}

	}

	return ptins;
}

tPProperty RemoveProperty(tPTag pTag, tPProperty pProp)
{
	tPProperty pp = pTag->Properties;

	if (pp == pProp)
	{
		pTag->Properties = pp->Next;
	}
	else
	{
		do
		{
			if (pp->Next == pProp)
			{
				pp->Next = pp->Next->Next;
				break;
			}
		}
		while (pp = pp->Next);
		return NULL;
	}
	return pProp;
}

tPProperty AddProperty(tPTag pTag, tPChar PropertyName)
{
	tPPropertyType ppt = NULL;
	tPProperty pp = NULL;

	pp = SearchProperty(pTag, PropertyName);
	if (pp == NULL)
	{
		if (pTag != NULL)
			ppt = SearchPropertyType(PropertyName, pTag->Type);

		if (ppt != NULL)
		{
			pp = malloctype(tProperty);
			if (pp != NULL)
			{
				memset(pp, 0, sizeof(tProperty));
				pp->Type = ppt;
				pp->Next = pTag->Properties;
				pTag->Properties = pp;
			}
		}
		else
		{
			if (WarnProperties)
			{
				if ((pTag != NULL) && (pTag->Type != NULL)
					&& (pTag->Type->Name != NULL))
				{
					fprintf(myout,
							"The property \"%s\" can't be created under \"%s\"!\n",
							PropertyName, pTag->Type->Name);
				}
				else
				{
					fprintf(myout,
							"The property \"%s\" can't be created under \"\"!\n",
							PropertyName);
				}
			}
		}
	}

	return pp;
}

tPProperty
AddPropertyValue(tPTag pTag, tPChar PropertyName, tPChar PropertyValue)
{
	tPProperty pp = AddProperty(pTag, PropertyName);

	if (pp != NULL)
	{
		pp->Value = PropertyValue;
	}
	else
		free(PropertyValue);
	return pp;
}

void
AddTreePropertyValue(tPTag pTag, tPChar PropertyName, tPChar PropertyValue,
					 char ProcessSubAfterHit, char ProcessNext)
{
	tPProperty pp;

	if ((pTag == NULL) || (PropertyName == NULL))
	{
		return;
	}

	pp = SearchProperty(pTag, PropertyName);

	if (pp == NULL)
		pp = AddPropertyValue(pTag, PropertyName, strdup(PropertyValue));

	if ((pp == NULL) || (ProcessSubAfterHit == true))
	{
		AddTreePropertyValue(pTag->SubTags, PropertyName, PropertyValue,
							 ProcessSubAfterHit, true);
	}

	if (ProcessNext)
		AddTreePropertyValue(pTag->Next, PropertyName, PropertyValue,
							 ProcessSubAfterHit, true);
}

char IsTrivialTree(int depth, tPTag pt)
{
	char ret = 0;

	if (pt != NULL)
	{
		ret += ((pt->TextIn == NULL) ||
				(strspn(pt->TextIn, " \t\n\r") != strlen(pt->TextIn))) ? 0 : 1;
		ret += (pt->SubTags == NULL) ? 0 : 1;
		ret += (IsTrivialTree(depth + 1, pt->SubTags) != 0) ? 0 : 1;
		ret += ((pt->SubTags == NULL) || (pt->SubTags->Next == NULL)) ? 0 : 1;
		if (depth > 1)
			ret += ((pt->TextAfter == NULL) ||
					(strspn(pt->TextAfter, " \t\n\r") !=
					 strlen(pt->TextAfter))) ? 0 : 1;
	}

	return (ret <= 1);
}

tPTag GetPrevTag(tPTag pt)
{
	tPTag ptp = pt->Parent->SubTags;

	if (ptp == pt)
		return NULL;
	while ((ptp != NULL) && (ptp->Next != pt))
		ptp = ptp->Next;
	return ptp;
}

void printTagFilePos(FILE * f, tPTag pt)
{
	tPProperty pp;

	pp = GetConfig(pt, "html-filepos");
	if ((pp != NULL) && (pp->Value != NULL))
		fprintf(f, "Pos in html: %s\n", pp->Value);
}

/******************************************************
*  Andere Funktionen                                  *
******************************************************/

long a2l(tPChar Value, long StdValue)
{
	long v;
	char *c = Value;

	/*
	 * Wenn ein Zeichen danach nicht lesbar ist, dann ist es ok
	 */

	if (Value != NULL)
	{
		errno = 0;
		v = strtol(Value, &c, 0);
	}
	else
		errno = -1;

	if (errno != 0)
	{
		v = StdValue;
	}
	if (debmodeorig == 2)
		fprintf(mylog, "a2l \"%s\" => %li\n", Value, v);
	return v;
}

double a2f(tPChar Value, double StdValue)
{
	double v;
	char *c = Value;

	/*
	 * Wenn ein Zeichen danach nicht lesbar ist, dann ist es ok
	 */

	if (Value != NULL)
	{
		errno = 0;
		v = strtod(Value, &c);
	}
	else
		errno = -1;

	if (errno != 0)
	{
		v = StdValue;
	}
	if (debmodeorig == 2)
		fprintf(mylog, "a2f \"%s\" => %f\n", Value, v);
	return v;
}

double ap2f(tPChar Value, double StdValue, double percentbase)
{
	double v;
	char *c = Value;

	/*
	 * Wenn ein Zeichen danach nicht lesbar ist, dann ist es ok 
	 */
	if (Value != NULL)
	{
		errno = 0;
		v = strtod(Value, &c);
	}
	else
		errno = -1;

	if (errno != 0)
	{
		v = StdValue;
	}
	else
	{
		if (*c == '%')
		{
			v = percentbase * v / 100;
		}
	}
	if (debmodeorig == 2)
		fprintf(myout, "ap2f \"%s\" => %f%c (pb = %f)\n", Value, v, *c,
				percentbase);

	return v;
}

void ReadTag(char read_until, tPPChar c)
{
	tPChar savefbuf = *fbuf;

	savefbuf = *fbuf;
	while ((**fbuf != read_until) && (**fbuf != '\0'))
	{
		(*fbuf)++;
	}
	if (debmodeorig)
		fprintf(mylog, "readed %i chars\n", ((*fbuf - savefbuf)));
	if (*fbuf != savefbuf)
	{
		if (*c == NULL)
		{
			*c = (char *) malloc((*fbuf - savefbuf) + 1);
			memcpy(*c, savefbuf, (*fbuf - savefbuf));
			*(*c + (*fbuf - savefbuf)) = '\0';
		}
		else
		{
			*c = (char *) realloc(*c, strlen(*c) + (*fbuf - savefbuf) + 1);
			*(*c + strlen(*c) + (*fbuf - savefbuf)) = '\0';
			memcpy(*c + strlen(*c), savefbuf, (*fbuf - savefbuf));
		}

		if (debmodeorig)
			fprintf(mylog, "String:%s\n", *c);
	}
	else
	{
		*c = NULL;
	}
	if (**fbuf == read_until)
		(*fbuf)++;
}

int strmulticmp(tPChar find, tPChar multi)
{
	int i;

	if ((find != NULL) && (*find != 0) && (multi != NULL))
	{
		while (*multi != 0)
		{
			if (*multi == '%')
			{
				i = 0;
				while ((DefTagGroups[i].Name != NULL)
					   && (strcasecmp(multi, DefTagGroups[i].Name) != 0))
					i++;
				if ((DefTagGroups[i].Name != NULL)
					&& (strmulticmp(find, DefTagGroups[i].MapTo) == 0))
				{
					return 0;
				}
			}

			if (strcasecmp(find, multi) == 0)
			{
				return 0;
			}

			multi += strlen(multi) + 1;
		}
	}
	return -1;
}

int isspacestr(char *s)
{
	if (s != NULL)
	{
		while (*s != 0)
		{
			if (!isspace(*s))
				return false;
			s++;
		}
	}

	return true;
}

void
ParseProperty(tPProperty * ppProp, tPChar Name, tPChar Value,
			  tPTagType pTagType)
{
	tPProperty pp;
	tPPropertyType ppt;

	if (Name)
	{
		do
		{
			ppt = SearchPropertyType(Name, pTagType);
			if ((ppt != NULL) && (ppt->MapTo != NULL) && (*ppt->MapTo == '@'))
			{
				fprintf(mylog,
						"Processing property \"%s\" like property ", Name);
				free(Name);
				Name = strdup(ppt->MapTo + 1);
				fprintf(mylog, "\"%s\"\n", Name);
			}
			else
				break;
		}
		while (true);

		if (ppt)
		{
			pp = malloctype(tProperty);
			pp->Type = ppt;
			pp->Value = Value;
			pp->Next = *ppProp;
			*ppProp = pp;
		}
		else
		{
			fprintf(myout, "\nProperty \"%s\" not found!!!\n", Name);

			if (Value != NULL)
			{
				fprintf(myout, "Value %s ", Value);
				free(Value);
			}

			if (pTagType->Name)
				fprintf(myout, "Tag %s", pTagType->Name);

			fprintf(myout, "\n\n");
		}
		free(Name);
	}
}

void ParseProperties(tPTag pTag)
{
	tPChar temps = pTag->FullTag;
	char inDeli = 0, SetSign = 0, danc = 0;	/* DoAsNormalChar */
	tPChar Name = NULL, Value = NULL;

	temps = strpbrk(temps, " \t\n\r");

	if (temps)
	{
		do
		{
			switch (*temps)
			{
			case '\'':
			case '\"':
				if (inDeli == 0)
				{
					inDeli = *temps;
					break;
				}
				else
				{
					if (inDeli == *temps)
					{
						inDeli = 0;
						if (Value == NULL)
							Value = strdup("");
						break;
					}
				}
				danc = 1;
				break;

			case '=':
				if (SetSign == 0)
				{
					SetSign = 1;
					break;
				}
				danc = 1;
				break;

			case ' ':
			case '\t':
			case '\n':
			case '\r':
			case '\0':
				if ((((SetSign == 1) && (Value != NULL)) ||
					 ((SetSign == 0) && (Value == NULL))) &&
					((inDeli == 0) || (*temps == '\0')))
				{
					SetSign = 0;
					if ((Value != NULL) && (*Value == '\0'))
					{
						free(Value);
						Value = NULL;
					}
					ParseProperty(&(pTag->Properties), Name, Value, pTag->Type);
					Value = Name = NULL;
					break;
				}
				if ((SetSign == 1) && (Value != NULL))
					danc = 1;
				break;

			default:
				danc = 1;
			}

			if (danc)
			{
				danc = 0;
				if (SetSign)
				{
					if (Value)
					{
						Value = (char *) realloc(Value, strlen(Value) + 2);
					}
					else
					{
						Value = (char *) malloc(2);
						*Value = '\0';
					}
					*(Value + strlen(Value) + 1) = 0;
					*(Value + strlen(Value)) = *temps;
				}
				else
				{
					if (Name)
					{
						Name = (char *) realloc(Name, strlen(Name) + 2);
					}
					else
					{
						Name = (char *) malloc(2);
						*Name = '\0';
					}
					*(Name + strlen(Name) + 1) = 0;
					*(Name + strlen(Name)) = *temps;
				}
			}
			temps++;
		}
		while (*(temps - 1) != '\0');
	}
	else
	{

	}

	ParseProperty(&(pTag->Properties), Name, Value, pTag->Type);
}

char FindStringInArray(tPChar * array, tPChar str)
{
	char i = 0;;
	while (*array != NULL)
	{
		if (strcasecmp(*array, str) == 0)
		{
			return i;
		}
		i++;
		array++;
	}

	return -1;
}

void ReadTree()
{
	tPChar activeTagString = NULL;
	tPChar activeTagName;
	tPChar BadClosedTags[200] = { NULL };
	tPTagType activeTagType;
	tPTag activeTag;
	int len, tspos, bctpos = 0, i;
	char preClosed;
	char tmpstr[20];

	fbufbegin = *fbuf;
	ReadTag('<', &activeTagString);
	free(activeTagString);

	while (**fbuf != '\0')
	{
		activeTagString = NULL;

		tagstartpos = *fbuf - fbufbegin;

		ReadTag('>', &activeTagString);
		if (activeTagString == NULL)
		{
			fprintf(myout, "Found empty Tag!\nPos %i\n", tagstartpos);
			ReadTag('<', &(TagStack[TagStackPos - 1]->TextIn));
		}
		else
		{
			fprintf(mylog, "Processing Tag \"%s\"\n", activeTagString);
			if (*activeTagString != '/')
			{
				/*
				 * Neues Tag aufnehmen
				 */

				if (*(activeTagString + strlen(activeTagString) - 1) == '/')
				{
					*(activeTagString + strlen(activeTagString) - 1) = '\0';
					preClosed = true;
				}
				else
					preClosed = false;

				if (strncmp(activeTagString, "!--", 3) == 0)
				{
					len = 3;
				}
				else
				{
					len =
						(strpbrk(activeTagString, " \t\n\r") - activeTagString);
					if (len == -(long) activeTagString)
						len = strlen(activeTagString);
				}

				activeTagName = (char *) malloc(len + 1);

				strncpy(activeTagName, activeTagString, len);
				*(activeTagName + len) = '\0';
				do
				{
					activeTagType = SearchTagType(NULL, activeTagName);
					if ((activeTagType != NULL) &&
						(activeTagType->MapTo != NULL) &&
						(*activeTagType->MapTo == '@'))
					{
						fprintf(myout,
								"Processing tag \"%s\" like tag ",
								activeTagName);
						free(activeTagName);
						activeTagName = strdup(activeTagType->MapTo + 1);
						fprintf(mylog, "\"%s\"\nPos %i\n", activeTagName,
								tagstartpos);
					}
					else
						break;
				}
				while (true);

				activeTag = NULL;

				if (activeTagType == NULL)
				{
					fprintf(myout,
							"Tag \"%s\" not found in internal database!!!\nPos %i\n",
							activeTagName, tagstartpos);

					fprintf(myout, "TagString: \"%si\"\n\n", activeTagString);

					ReadTag('<', &(TagStack[TagStackPos - 1]->TextIn));
				}
				else
				{
					if ((activeTagType->MapTo != NULL)
						&& (*(activeTagType->MapTo) == '\0'))
					{
						ReadTag('<', &(TagStack[TagStackPos - 1]->TextIn));
					}
					else
					{
						i = 0;
						while ((TagStackPos > 0) &&
							   (TagStack[TagStackPos - 1]->Type->
								ValidSubTags != NULL)
							   &&
							   (strmulticmp
								(activeTagName,
								 TagStack[TagStackPos -
										  1]->Type->ValidSubTags) != 0))
						{
							if ((activeTagType->CloseOtherTags !=
								 NULL)
								&&
								(strmulticmp
								 (TagStack[TagStackPos - 1]->Type->
								  Name, activeTagType->CloseOtherTags) == 0))
							{
								fprintf(myout,
										"Tag \"%s\" not closed!!!\nPos %i\n",
										TagStack[TagStackPos - 1]->FullTag,
										tagstartpos);
								TagStackPos--;
							}
							else
							{
								fprintf(myout,
										"Tag \"%s\" is not allowed as child of \"%s\"!!!\nPos %i\n",
										activeTagName,
										TagStack[TagStackPos -
												 1]->Type->Name, tagstartpos);
								ReadTag('<',
										&(TagStack[TagStackPos - 1]->TextIn));
								i = 1;
								break;
							}
						}

						if (i == 0)
						{
							if (activeTagType->Cascade != Yes)
							{
								tspos = TagStackPos;
								while (tspos > 0)
								{
									tspos--;
									if (TagStack[tspos] != NULL)
									{
										if (activeTagType ==
											TagStack[tspos]->Type)
										{
											if ((activeTagType->Cascade ==
												 No)
												||
												((activeTagType->Cascade ==
												  MayBe)
												 && (TagStackPos == tspos + 1)))
											{
												fprintf
													(myout,
													 "Tag \"%s\" not or incorrect closed!!!\nPos %i\n",
													 TagStack[tspos]->
													 FullTag, tagstartpos);
												TagStackPos = tspos;
												break;
											}
										}
									}
								}
							}

							if ((TagStackPos > 0)
								&& (TagStack[TagStackPos - 1] != NULL))
							{
								while (strmulticmp
									   (TagStack[TagStackPos - 1]->
										Type->Name,
										activeTagType->CloseOtherTags) == 0)
								{
									fprintf(myout,
											"Tag \"%s\" not closed!!!\nPos %i\n",
											TagStack[TagStackPos
													 - 1]->FullTag,
											tagstartpos);
									TagStackPos--;
								}
							}
							activeTag = malloctype(tTag);
							memset(activeTag, 0, sizeof(tTag));
							activeTag->Type = activeTagType;
							activeTag->FullTag = activeTagString;
							fprintf(mylog, "open Tag %s\n", activeTagString);
							if (TagStackPos > 0)
							{
								activeTag->Parent = TagStack[TagStackPos - 1];
								if (TagStack[TagStackPos - 1]->
									LastSubTag != NULL)
								{
									TagStack[TagStackPos -
											 1]->LastSubTag->Next = activeTag;
								}
								else
								{
									TagStack[TagStackPos -
											 1]->SubTags = activeTag;
								}
								TagStack[TagStackPos -
										 1]->LastSubTag = activeTag;
							}
							else
							{
								if (ROOT == NULL)
								{
									ROOT = activeTag;
								}
								else
								{
									ROOT->Next = activeTag;
								}
							}

							if (activeTag->Type->ConvertProperties == Yes)
								ParseProperties(activeTag);

							TagStack[TagStackPos] = activeTag;
							if ((activeTag->Type->EndTag != No) && (!preClosed))
							{
								TagStackPos++;
								ReadTag('<', &(activeTag->TextIn));
							}
							else
							{
								ReadTag('<', &(activeTag->TextAfter));
							}
						}
					}
				}
				free(activeTagName);
				sprintf(tmpstr, "%i", tagstartpos);
				AddPropertyValue(activeTag, "html-FilePos", strdup(tmpstr));
			}
			else
			{
				/*
				 * delete old tag
				 */
				len = strcspn(activeTagString + 1, " \t\n\r");
				activeTagName = (char *) malloc(len + 1);
				strncpy(activeTagName, activeTagString + 1, len);
				*(activeTagName + len) = '\0';

				activeTagType = SearchTagType(NULL, activeTagName);
				if ((activeTagType != NULL) &&
					(activeTagType->MapTo != NULL) &&
					(*(activeTagType->MapTo) == '\0'))
				{
					ReadTag('<', &(TagStack[TagStackPos - 1]->TextIn));
				}
				else
				{
					while ((activeTagType != NULL) &&
						   (activeTagType->MapTo != NULL) &&
						   (*(activeTagType->MapTo) == '@'))
					{
						free(activeTagName);
						len = strlen(activeTagType->MapTo) - 1;
						activeTagName = (char *) malloc(len + 1);
						strncpy(activeTagName, activeTagType->MapTo + 1, len);
						*(activeTagName + len) = '\0';

						activeTagType = SearchTagType(NULL, activeTagName);
					}

					tspos = TagStackPos;
					while (tspos > 0)
					{
						tspos--;
						if (TagStack[tspos] != NULL)
						{
							if (strcasecmp
								(activeTagName,
								 TagStack[tspos]->Type->Name) == 0)
							{
								fprintf(mylog,
										"\t found tag \"%s\"...\n",
										TagStack[tspos]->FullTag);

								while ((tspos > 0)
									   &&
									   ((i =
										 FindStringInArray
										 (&BadClosedTags[0],
										  TagStack[tspos -
												   1]->Type->Name)) >= 0))
								{
									fprintf(myout,
											"Tag \"%s\" incorrect closed!!!\nPos %i\n",
											BadClosedTags[i], tagstartpos);
									free(BadClosedTags[i]);
									for (; i < bctpos; i++)
									{
										BadClosedTags[i] = BadClosedTags[i + 1];
									}
									bctpos--;
									tspos--;
								}

								TagStackPos = tspos;
								ReadTag('<',
										&(TagStack[TagStackPos]->TextAfter));
								break;
							}
							else
							{
								if (TagStack[tspos]->Type->EndTag == MayBe)
								{
									fprintf(myout,
											"Tag \"%s\" not closed!!!\nPos %i\n",
											TagStack[tspos]->FullTag,
											tagstartpos);
								}
								else
								{
									if (TagStack[tspos]->Type->EndTag == Yes)
									{
										if ((i =
											 FindStringInArray
											 (&BadClosedTags
											  [0],
											  TagStack[tspos]->
											  Type->Name)) >= 0)
										{
											fprintf(myout,
													"Tag \"%s\" incorrect closed!!!\nPos %i\n",
													BadClosedTags[i],
													tagstartpos);
											free(BadClosedTags[i]);
											for (; i < bctpos; i++)
											{
												BadClosedTags
													[i] = BadClosedTags[i + 1];
											}
											bctpos--;
										}
										else
										{
											fprintf(myout,
													"Tag \"%s\" incorrect closed!!!\nPos %i\n",
													activeTagString,
													tagstartpos);
											BadClosedTags[bctpos] =
												strdup(activeTagName);
											bctpos++;
											while (TagStack[TagStackPos] ==
												   NULL)
											{
												TagStackPos--;
											}
											ReadTag('<',
													&TagStack
													[TagStackPos]->TextAfter);
											break;
										}
									}
								}
							}
						}
					}
				}
				free(activeTagName);
				if (activeTagString != NULL)
				{

					if (strcasecmp(activeTagString, "/html") == 0)
					{
						free(activeTagString);
						fprintf(mylog, "returning after /html\n");
						break;
					}

					free(activeTagString);
				}
			}
		}
		fprintf(mylog, "bctpos %i\n", bctpos);
	}
}

void ReadDaten(FILE * f)
{
	long flen;
	tPChar fbuf1, fbuf2;

	fbuf = &fbuf1;
	fseek(f, 0, SEEK_END);
	flen = ftell(f);
	fseek(f, 0, SEEK_SET);
	fbuf2 = fbuf1 = (char *) malloc(flen + 5);	/* one additional byte should also work, but for some reasons. it fails sometimes ;-( */
	fread(fbuf1, flen, 1, f);
	*(fbuf1 + flen) = '\0';
	fprintf(mylog, "%li\n", flen);
	fbuflen = strlen(fbuf1);
	ReadTree();
	fbufbegin = NULL;
	free(fbuf2);
}

void ProcessSpecialChars(tPPChar str)
{
	int i = 0, diff;
	tPChar s = *str, s2;

	while (*s != 0)
	{
		i = 0;
		while (DefSpecialChar[i].Name != NULL)
		{
			s2 = NULL;
			while (strncmp
				   (s, DefSpecialChar[i].Name,
					strlen(DefSpecialChar[i].Name)) == 0)
			{
				if ((s2 == NULL) && (DefSpecialChar[i].MapToBefore != NULL))
				{
					diff = s - *str;
					s2 = strdup(s);
					*str =
						realloc(*str,
								(strlen(*str) +
								 strlen(DefSpecialChar[i].MapToBefore)) + 1);
					s = *str + diff;
					/*
					 * copying the trailing text... 
					 */
					memcpy(s +
						   strlen(DefSpecialChar[i].MapToBefore),
						   s2, strlen(s2) + 1);
					free(s2);
					/*
					 * copying the replaced text 
					 */
					memcpy(s, DefSpecialChar[i].MapToBefore,
						   strlen(DefSpecialChar[i].MapToBefore));
					s += strlen(DefSpecialChar[i].MapToBefore);
				}

				if (DefSpecialChar[i].MapTo != NULL)
				{
					diff = s - *str;
					s2 = strdup(s + strlen(DefSpecialChar[i].Name));
					*str =
						realloc(*str,
								(strlen(*str) -
								 strlen(DefSpecialChar[i].Name) +
								 strlen(DefSpecialChar[i].MapTo)) + 1);
					s = *str + diff;
					/*
					 * copying the trailing text...
					 */
					memcpy(s + strlen(DefSpecialChar[i].MapTo), s2,
						   strlen(s2) + 1);
					free(s2);
					/*
					 * copying the replaced text
					 */
					memcpy(s, DefSpecialChar[i].MapTo,
						   strlen(DefSpecialChar[i].MapTo));
					s += strlen(DefSpecialChar[i].MapTo);
				}
				else
					s += strlen(DefSpecialChar[i].Name);
			}
			if ((s2 != NULL) && (DefSpecialChar[i].MapToAfter != NULL))
			{
				diff = s - *str;
				s2 = strdup(s);
				*str =
					realloc(*str,
							(strlen(*str) +
							 strlen(DefSpecialChar[i].MapToAfter)) + 1);
				s = *str + diff;
				/*
				 * copying the trailing text...
				 */
				memcpy(s + strlen(DefSpecialChar[i].MapToAfter), s2,
					   strlen(s2) + 1);
				free(s2);
				/*
				 * copying the replaced text
				 */
				memcpy(s, DefSpecialChar[i].MapToAfter,
					   strlen(DefSpecialChar[i].MapToAfter));
				s += strlen(DefSpecialChar[i].MapToAfter);
			}
			if (s2 != NULL)
				break;
			i++;
		}
		if (s2 == NULL)
			s++;
	}
}

void AddPreDefProperties(tPTag pt)
{
	tPPreDefProperty ppdp = &DefPreDefProperties[0];

	while (ppdp->Tag != NULL)
	{
		if (strcasecmp(pt->Type->Name, ppdp->Tag) == 0)
		{
			if (ppdp->Value != NULL)
			{
				AddPropertyValue(pt, ppdp->Property, strdup(ppdp->Value));
			}
			else
			{
				AddProperty(pt, ppdp->Property);
			}
		}
		ppdp++;
	}
}

void ConvertTree(tPTag pt)
{

	tPProperty pp;

	while (pt != NULL)
	{
		if (debmodeorig)
			fprintf(mylog, "<%s>\n", NullString(pt->FullTag));

		pp = pt->Properties;

		if (pt->Type->pProcessTagBefore != NULL)
		{
			(*(pt->Type->pProcessTagBefore)) (pt);
		}

		AddPreDefProperties(pt);

		while (pp != NULL)
		{
			if (debmodeorig)
				fprintf(mylog, "\"%s\" : \"%s\"@%8p\n",
						NullString(pp->Type->Name),
						NullString(pp->Type->Name),
						NullString(pp->Value), NullString(pp->Value));

			if (pp->Type->pConvertPropertyType != NULL)
			{
				if ((*(pp->Type->pConvertPropertyType)) (pt, pp) == false)
				{
					fprintf(myout,
							"Property \"%s\" (Value \"%s\") within tag \"%s\" was not correct converted!\n",
							NullString(pp->Type->Name),
							NullString(pp->Value), NullString(pt->FullTag));
				}
				if (debmodeorig)
					fprintf(mylog, "%*s : \"%s\"@%8p\n",
							strlen(NullString(pp->Type->Name)) + 2,
							"", NullString(pp->Value), NullString(pp->Value));
			}

			pp = pp->Next;
		}

		if (debmodeorig)
			fprintf(mylog, "%s", NullString(pt->TextIn));

		ConvertTree(pt->SubTags);

		if (debmodeorig)
			fprintf(mylog, "%s", NullString(pt->TextAfter));

		if (pt->Type->pProcessTag != NULL)
		{
			(*(pt->Type->pProcessTag)) (pt);
		}

		pt = pt->Next;
	}

	return;
}

void WriteTree(FILE * f, tPTag pt, int IndentDepth)
{
	tPTag pts;
	tPProperty pp;

	if (pt != NULL)
	{
		if (pt->Type->MapTo != NULL)
		{
			if ((pt->Type->ConvertProperties == Yes) || (pt->FullTag == NULL))
			{
				fprintf(f, "<%s", pt->Type->MapTo);
				pp = pt->Properties;

				while (pp != NULL)
				{
					if (pp->Type->MapTo != NULL)
					{
						if (*(pp->Type->MapTo) != '\0')
							fprintf(f, " %s", pp->Type->MapTo);

						if (pp->Value != NULL)
						{
							ProcessSpecialChars(&pp->Value);

							if (*(pp->Type->MapTo) != '\0')
								fprintf(f, "=\"");

							fprintf(f, "%s", pp->Value);

							if (*(pp->Type->MapTo) != '\0')
								fprintf(f, "\"");
						}
					}

					pp = pp->Next;
				}

				if (pt->Type->CloseXMLTag == MayBe)
				{
					fprintf(f, "/");
				}
			}
			else
				fprintf(f, "<%s", pt->FullTag);

			fprintf(f, ">");
		}

		if (pt->Type->CloseXMLTag == Yes)
		{
			if (pt->TextIn != NULL)
			{
				ProcessSpecialChars(&pt->TextIn);

				fprintf(f, "%s", pt->TextIn);
			}

			pts = pt->SubTags;

			while (pts != NULL)
			{
				WriteTree(f, pts, IndentDepth + 1);
				pts = pts->Next;
			}

			if (pt->Type->MapTo != NULL)
			{
				fprintf(f, "</%s>", pt->Type->MapTo);
			}
		}

		if (pt->TextAfter != NULL)
		{
			ProcessSpecialChars(&pt->TextAfter);

			fprintf(f, "%s", pt->TextAfter);
		}
	}
}

void WriteDaten(FILE * f)
{
	tPTag pt = ROOT;

	fprintf(myout, "converting special chars and writing out...");
	fflush(myout);

	while (pt != NULL)
	{
		WriteTree(f, pt, 0);
		pt = pt->Next;
	}

	fprintf(f, "\r\n");

	fprintf(myout, " done\n");
}

void FreeProperty(tPPProperty ppProperty)
{
	if (*ppProperty)
	{
		free((*ppProperty)->Value);
		FreeProperty(&(*ppProperty)->Next);
		free((*ppProperty));
	}
}

void FreeTag(tPPTag ppTag)
{
	if ((*ppTag))
	{
		free((*ppTag)->FullTag);
		FreeProperty(&(*ppTag)->Properties);
		free((*ppTag)->TextIn);
		FreeTag(&(*ppTag)->SubTags);
		free((*ppTag)->TextAfter);
		FreeTag(&(*ppTag)->Next);
		free((*ppTag));
	}
}

void FreeDaten()
{
	fprintf(myout, "freeing memory...");
	fflush(myout);

	FreeTag(&ROOT);

	fprintf(myout, " done\n");
}

tPChar getIndent(tPTag pTag)
{
	tPTag pt = GetPrevTag(pTag);
	tPChar s, s2 = NULL;
	int i = 0;
	char cont = true;

	if (pt == NULL)
	{
		s = pTag->Parent->TextIn;
	}
	else
	{
		s = pt->TextAfter;
	}

	if (s != NULL)
	{
		s2 = s;
		while (cont == true)
		{
			switch (*(s + i))
			{
			case ' ':
			case '\t':
				break;

			case '\x0a':		/* also for DOS-formated files */
			case '\x0d':
				s2 = s + i + 1;
				break;

			default:
				cont = false;
			}
			if (cont == true)
				i++;
		}
		i = i - (s2 - s);
		if (i > 0)
		{
			s = malloc(i + 1);
			memcpy(s, s2, i);
			*(s + i) = '\0';
		}
		else
			s = NULL;
	}
	return s;
}

tPProperty AddPropertySeparator(tPTag pTag)
{
	tPProperty ppt;
	tPChar s, s2 = NULL;

	s = getIndent(pTag);
	if (s != NULL)
	{
		s2 = malloc(strlen(s) + 6);
		sprintf(s2, "\n%s    ", s);
		free(s);
	}
	ppt = AddPropertyValue(pTag, "LIST-SEPARATOR", s2);
	return ppt;
}

int main(int argc, char **argv)
{
	FILE *fh, *ff;

	/*
	 * char OutputName[80];
	 */
	myout = stdout;
	mylog = myout;

	InitROOT();
	debmodeorig = (argc == 4) ? 1 : 0;
	if (debmodeorig)
	{
		if (strcmp(argv[3], "debug2") == 0)
			debmodeorig++;
	}
	else
		mylog = fopen(NullDevice, "w");

	if (mylog == NULL)
		mylog = myout;

	if (debmodeorig)
	{
		fprintf(mylog,
				"starting html2fo - DebugMode %i\nparameter count: %i\n",
				debmodeorig, argc);
	}
	if ((argc < 3) || (argc > 4))
	{
		fprintf(myout,
				"html2fo (HTML-file) (FO-file) [debug]\ndebug prints debugging messages to stderr.\n");
		/*
		 * DisplayHelp();
		 */
	}
	else
	{
		/*
		 * Daten
		 */
		if ((fh = fopen(argv[1], "r")) == NULL)
		{
			fprintf(myout, "file %s is not readable!\n", argv[1]);
			return 1;
		}

		strcpy(myinputfile, argv[1]);
		ReadDaten(fh);
		fclose(fh);

		ConvertTree(ROOT);

		if ((ff = fopen(argv[2], "w")) == NULL)
		{
			fprintf(myout, "file %s seems to be write protected!\n", argv[2]);
			FreeDaten();
			return 1;
		}
		WriteDaten(ff);

		fclose(ff);
		FreeDaten();
	}

	fprintf(myout, "\nall is done for file %s\n", myinputfile);
	if (mylog != myout)
		fclose(mylog);
	return 0;
}

