#include "html2fo.h"
#include "structs.h"

tPChar HTML_Title = NULL;
tPChar HTML_Footer = NULL;

/******************************************************
 * conversion funktions for property data             *
 *****************************************************/

char cptSetNull(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;

	free(*Value);
	*Value = NULL;
	return true;
}

char cptFontSize(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;
	int v, count, v2, v3;
	char outc[8], c, *mv;

	/* it is ok if the sign after my value is not readable */
	mv = *Value;
	if ((*mv == '-') || (*mv == '+'))
		mv++;

	count = sscanf(mv, "%i%c", &v, &c);

	if (**Value == '-')
	{
		v = 2 - v;
	}

	if (**Value == '+')
	{
		v = 2 + v;
	}

	if (v < 0)
		v = 0;

	if (count == 1)
	{
		switch (v)
		{
		case 0:
			v2 = 7;
			break;
		case 1:
			v2 = 8;
			break;
		case 2:
			v2 = 10;
			break;
		case 3:
			v2 = 13;
			break;
		case 4:
			v2 = 16;
			break;
		case 5:
			v2 = 20;
			break;
		default:
			v2 = v * 5 - 5;		/* 6=25 7=30 8=35... */
		}
		v3 = v2 + 3;
		sprintf(outc, "%ipt", v2);
		fprintf(mylog, "Old FontSize %s\n", *Value);

/*		free(*Value);
		*Value=strdup(outc);
		fprintf(mylog,"New FontSize %s\n",*Value);
*/
		fprintf(mylog, "New FontSize %s\n",
				AddPropertyValue(pTag, "STYLE-font-size", strdup(outc))->Value);

		sprintf(outc, "%ipt", v3);
		fprintf(mylog, "LineHeight %s\n",
				AddPropertyValue(pTag, "XML-LINE-HEIGHT", strdup(outc))->Value);
		return true;
	}
	else
	{
		fprintf(mylog, "No FontSize!\n");
		return false;
	}
}

char cptPoint(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;

	if (*Value != NULL)
	{
		if ((strlen(*Value) > 1)
			&& ((strcasecmp(*Value + strlen(*Value) - 2, "pt") == 0)
				|| (strcasecmp(*Value + strlen(*Value) - 2, "px") == 0)))
		{						/* data -> append a "pt" */
			strcpy((*Value) + strlen(*Value) - 2, "pt");
		}
		else
		{						/* data -> append a "pt" */
			*Value = (char *) realloc(*Value, strlen(*Value) + 3);
			strcpy((*Value) + strlen(*Value), "pt");
		}
	}
	else
	{							/* no data -> = "1pt" */
		*Value = (tPChar) strdup("1pt");
	}
	return true;
}

char cptPadding(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;
	char oldwarn = WarnProperties;

	cptPoint(pTag, pProp);

	WarnProperties = false;

	if (strcmp(*Value, "0pt") != 0)
	{
		AddTreePropertyValue(pTag->SubTags, "XML-MARGIN-TOP", *Value,
							 false, true);
		AddTreePropertyValue(pTag->SubTags, "XML-MARGIN-BOTTOM", *Value,
							 false, true);
		AddTreePropertyValue(pTag->SubTags, "XML-MARGIN-LEFT", *Value,
							 false, true);
		AddTreePropertyValue(pTag->SubTags, "XML-MARGIN-RIGHT", *Value,
							 false, true);
	}

	WarnProperties = oldwarn;

	return true;
}

char cptBorder(tPTag pTag, tPProperty pProp)
{
	tPPChar Value;
	char ret;

	ret = cptPoint(pTag, pProp);

	Value = &pProp->Value;

	if ((ret) && (*Value != NULL) && (strcmp(*Value, "0pt") != 0))	/* parameter >1pt specified */
	{
		AddPropertyValue(pTag, "XML-BORDER-WIDTH", strdup(*Value));
		AddPropertyValue(pTag, "XML-BORDER-COLOR", strdup("black"));
		AddPropertyValue(pTag, "XML-BORDER-STYLE", strdup("solid"));
	}

	return ret;
}

char cptTableBorder(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;
	tPProperty pp;

	if (*Value == NULL)			/* no parameter specified */
	{
		pp = SearchProperty(pTag, "CELLSPACING");
		if (pp != NULL)			/* but may be here */
		{
			if (pp->Value != NULL)
			{
				*Value = strdup(pp->Value);
				memblocks++;
			}
		}
		else
			*Value = strdup("1pt");
	}

	return cptPoint(pTag, pProp);
}

char cptGenericMapping(tPTag pTag, tPProperty pProp, tPChar Map[][2])
{
	tPPChar Value = &pProp->Value;
	int i;

	for (i = 0; Map[i][0] != NULL; i++)
	{
		if (strcasecmp(Map[i][0], *Value) == 0)
		{
			free(*Value);
			*Value = strdup(Map[i][1]);
			fprintf(mylog, "map %s -> %s\n", Map[i][0], Map[i][1]);
			return true;
		}
	}

	return false;
}

char cptAlign(tPTag pTag, tPProperty pProp)
{
	tPChar Map[][2] = {
		{"LEFT", "start"},
		{"CENTER", "center"},
		{"MIDDLE", "center"},
		{"RIGHT", "end"},
		{NULL}
	};

	return cptGenericMapping(pTag, pProp, Map);
}

char cptGenericBreak(tPTag pTag, tPProperty pProp)
{
	tPChar Map[][2] = {
		{"ALWAYS", "page"},
		{NULL}
	};

	return cptGenericMapping(pTag, pProp, Map);
}

char cptParseProperties(tPTag pTag, tPProperty pProp)
{
	/* temp string */
	tPChar temps = pProp->Value;

	/* within Delimiter? */
	char inDeli = 0;

	/* within name or parameter (while parsing) */
	char SetSign = 0;

	/* danc = DoAsNormalChar */
	char danc = 0;

	/* name of style property */
	tPChar Name = NULL;

	/* parameter of style property */
	tPChar ParValue = NULL;

	tPProperty pp = NULL, origpp = pTag->Properties;

	if (temps)
	{
		do
		{
			switch (*temps)
			{
			case '\'':
			case '\"':
				if (inDeli == 0)
				{
					inDeli = *temps;
					break;
				}
				else
				{
					if (inDeli == *temps)
					{
						inDeli = 0;
						if (ParValue == NULL)
							ParValue = strdup("");
						break;
					}
				}
				danc = 1;
				break;

			case ':':
			case '=':
				if (SetSign == 0)
				{
					SetSign = 1;
					while (*(temps + 1) == ' ')
						temps++;
					break;
				}
				danc = 1;
				break;

			case ';':
			case '\0':
				if ((((SetSign == 1) && (ParValue != NULL)) ||
					 ((SetSign == 0) && (ParValue == NULL))) &&
					((inDeli == 0) || (*temps == '\0')))
				{
					SetSign = 0;
					if (Name != NULL)
						ParseProperty(&(pTag->Properties), Name, ParValue,
									  pTag->Type);
					while ((*temps != '\0') && (isspace(*(temps + 1))))
					{
						temps++;
					}

					Name = NULL;
					ParValue = NULL;
					break;
				}
				if ((SetSign == 1) && (ParValue != NULL))
					danc = 1;
				break;

			default:
				danc = 1;
			}

			if (danc)
			{
				danc = 0;
				if (SetSign)
				{
					if (ParValue)
					{
						ParValue =
							(char *) realloc(ParValue, strlen(ParValue) + 2);
					}
					else
					{
						ParValue = (char *) malloc(2);
						*ParValue = '\0';
					}
					*(ParValue + strlen(ParValue) + 1) = 0;
					*(ParValue + strlen(ParValue)) = *temps;
				}
				else
				{
					if (!Name)
					{
						Name = malloc(strlen(pProp->Type->Name) + 2);
						sprintf(Name, "%s-", pProp->Type->Name);
					}
					Name = (char *) realloc(Name, strlen(Name) + 2);
					*(Name + strlen(Name) + 1) = 0;
					*(Name + strlen(Name)) = *temps;
				}
			}
			temps++;
		}
		while (*(temps - 1) != '\0');
	}

	ParseProperty(&(pTag->Properties), Name, ParValue, pTag->Type);

	pp = pTag->Properties;

	while (pp != origpp)
	{
		if (pp->Type->pConvertPropertyType != NULL)
		{
			if ((*(pp->Type->pConvertPropertyType)) (pTag, pp) == false)
			{
				fprintf(myout,
						"Property \"%s\" (Value \"%s\") within tag \"%s\" was not correct converted!\n",
						NullString(pp->Type->Name),
						NullString(pp->Value), NullString(pTag->FullTag));
			}
		}

		pp = pp->Next;
	}
	return true;
}

char cptLinkName(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;

	pTag->Type = SearchTagType(NULL, "XML-LINK-REFERENCE");
	return true;
}

char cptLink(tPTag pTag, tPProperty pProp)
{
	tPPChar Value = &pProp->Value;
	char *c, *val;
	char *target = NULL, *file = NULL, *ext = NULL, *base = NULL, *path = NULL;

	val = strdup(*Value);
	c = strrchr(val, '#');
	if (c != NULL)
	{
		*c = '\0';
		c++;
		target = strdup(c);
	}

	c = strrchr(val, '/');
	if (c != NULL)
	{
		c++;
		if (strlen(c) > 0)
			file = strdup(c);
		*c = '\0';
		path = strdup(val);
	}
	else
	{
		c = strrchr(val, '.');
		if (c != NULL)
		{
			target = strdup(val);
			*val = '\0';
		}
		else
		{
			if (strlen(val) > 0)
				file = strdup(val);
		}
	}

	if (file != NULL)
	{
		c = strrchr(file, '.');
		if (c != NULL)
		{
			*c = '\0';
			c++;
			ext = strdup(c);
			base = strdup(file);
			*(c - 1) = '.';
		}
		else
			base = strdup(file);
	}

	if ((path == NULL) && (file == NULL) && (target != NULL))
	{
		free(*Value);
		*Value = target;
		pProp->Type =
			SearchPropertyType("XML-INTERNAL-DESTINATION", pTag->Type);
		target = NULL;
	}

	if ((path == NULL) &&
		(base != NULL) &&
		((ext != NULL) &&
		 ((strcasecmp(ext, "htm") == 0) ||
		  (strcasecmp(ext, "html") == 0)) &&
		 (strcmp(file, myinputfile) != 0)) && (target == NULL))
	{
		free(*Value);
		*Value = malloc(strlen(base) + 5);
		sprintf(*Value, "%s%s", base, ".pdf");
	}

	if (target != NULL)
		free(target);

	if (file != NULL)
		free(file);

	if (ext != NULL)
		free(ext);

	if (base != NULL)
		free(base);

	if (path != NULL)
		free(path);

	free(val);

	return true;
}


/******************************************************
 *  tag-specific convertion functions                 *
 ******************************************************/

char InitROOT()
{
	ROOT = malloctype(tTag);
	memset(ROOT, 0, sizeof(tTag));
	ROOT->Type = SearchTagType(NULL, "XML-?");
	AddPropertyValue(ROOT, "XML-?-?", NULL);
	AddPropertyValue(ROOT, "XML-?-ENCODING", strdup("ISO-8859-1"));
	AddPropertyValue(ROOT, "XML-?-VERSION", strdup("1.0"));
	ROOT->TextAfter = strdup("\r\n");
	return 1;
}

char SetupParagraph(int depth, tPTag pt)
{
	char ret = false;
	tPTag pts;

	if (pt != NULL)
	{
		if (strmulticmp(pt->Type->Name, TG_BLOCK) == 0)
			return false;

		if ((pt->TextIn != NULL) &&
			(strspn(pt->TextIn, " \t\n\r") != strlen(pt->TextIn)))
		{
			ret = true;
		}

		pts = pt->SubTags;
		while (pts != NULL)
		{
			if ((strmulticmp(pts->Type->Name, TG_BLOCK) != 0)
				&& (strmulticmp(pts->Type->Name, TG_INLINE) == 0))
			{
				ret = true;
			}
			else
				ret = ret || SetupParagraph(depth + 1, pts);

			if ((pts->TextAfter != NULL) &&
				(strspn(pts->TextAfter, " \t\n\r") != strlen(pts->TextAfter)))
			{
				ret = true;
			}

			pts = pts->Next;
		}


		if (ret == true)
		{
			if ((depth > 0) && (pt->Type->MapTo == NULL))
			{
				return true;
			}
			else
			{
				if ((strcmp(pt->Type->Name, "IMG") != 0) &&
					((strcmp(pt->Type->Name, "A") != 0) ||
					 (isspacestr(pt->TextIn) != true) ||
					 (pt->SubTags == NULL) ||
					 (strcmp(pt->SubTags->Type->Name, "IMG") != 0) ||
					 (isspacestr(pt->SubTags->TextAfter) != true) ||
					 (pt->SubTags->Next != NULL)))
				{
					InsertAsSubTag(pt, "P");
				}
				else
					return true;
			}
		}

	}

	return false;
}

tPTag SetupFont(int depth,
				tPTag pt,
				tPChar Font,
				tPChar Size,
				tPChar Color,
				tPChar LineHeight,
				char Bold,
				char Italic,
				char Underline, tPTag InParagraph, tPTag TrivialInlineOrBlock)
{
	tPTag is, pttmp;
	tPProperty tpp;

	if (pt != NULL)
	{
		fprintf(mylog, "%*s%s\n", depth * 4, "", pt->Type->Name);

		if (strmulticmp(pt->Type->Name, TG_PARAGRAPH) == 0)
		{
			/* I am in a paragraph !!! */
			InParagraph = pt;

			AddPropertyValue(pt, "XML-KEEP-TOGETHER", strdup("always"));
			AddPropertyValue(pt, "XML-SPACE-AFTER", strdup("1.5pt"));
			AddPropertyValue(pt, "XML-SPACE-BEFORE", strdup("1.5pt"));

/*			if (((pt->TextIn==NULL)||(isspacestr(pt->TextIn)!=0))&&(pt->SubTags==NULL))
			{
				if (pt->TextIn!=NULL)
					free(pt->TextIn);
				pt->TextIn=strdup("\r\n");
				AddPropertyValue(pt,"XML-WHITE-SPACE-COLLAPSE","false");
			}
*/
			if (IsTrivialTree(0, pt))
			{
				TrivialInlineOrBlock = pt;
			}
			else
			{
				TrivialInlineOrBlock = NULL;
			}
		}
		else
		{
			if (strmulticmp(pt->Type->Name, TG_BLOCK_FB) == 0)
			{
				InParagraph = NULL;
				TrivialInlineOrBlock = NULL;
			}
			else
			{
				if ((pt->Type == SearchTagType(pt->Parent, "XML-INLINE"))
					&& (IsTrivialTree(0, pt)))
				{
					TrivialInlineOrBlock = pt;
				}
				else
				{
					if ((InParagraph != NULL)
						&& (TrivialInlineOrBlock == NULL)
						&& (pt->Type->EndTag != No))
					{
						if (IsTrivialTree(0, pt))
						{
							is = InsertToSubTag(pt, "XML-INLINE");
							if (is != NULL)
							{
								TrivialInlineOrBlock = is;
								pt = is;
							}
						}
						else
						{
							if ((pt->TextIn != NULL)
								&& (isspacestr(pt->TextIn) == false))
							{
								AddSubTagFirstIncludeText(pt, "XML-INLINE");
							}
						}
					}

				}
			}
		}

		if (strmulticmp(pt->Type->Name, TG_FONTTAGS) == 0)
		{
			tpp = SearchProperty(pt, "XML-FONT-FACE");
			if (tpp != NULL)
			{
				Font = tpp->Value;
			}

			tpp = SearchProperty(pt, "STYLE-font-size");
			if (tpp != NULL)
			{
				Size = tpp->Value;
			}

			tpp = SearchProperty(pt, "STYLE-color");
			if (tpp != NULL)
			{
				Color = tpp->Value;
			}

			tpp = SearchProperty(pt, "XML-LINE-HEIGHT");
			if (tpp != NULL)
			{
				LineHeight = tpp->Value;
			}
		}
		else
		{
			if (pt->Type == SearchTagType(NULL, "B"))
			{
				Bold = true;
			}
			else
			{
				if (pt->Type == SearchTagType(NULL, "I"))
				{
					Italic = true;
				}
				else
				{
					if (pt->Type == SearchTagType(NULL, "U"))
					{
						Underline = true;
					}
				}
			}
		}
		pttmp = pt;

		if ((InParagraph != NULL) && (TrivialInlineOrBlock == NULL)
			&& (pt->TextIn != NULL) && (isspacestr(pt->TextIn) == false))
		{
			AddSubTagFirstIncludeText(pt, "XML-INLINE");
		}

		pt = pt->SubTags;
		if (pt != NULL)
		{
			while (pt)
			{
				pt = SetupFont(depth + 1, pt, Font, Size, Color,
							   LineHeight, Bold, Italic, Underline,
							   InParagraph, TrivialInlineOrBlock);

				if ((InParagraph != NULL) && (TrivialInlineOrBlock == NULL)
					&& (pt->TextAfter != NULL)
					&& (isspacestr(pt->TextAfter) == false))
				{
					is = AddTagAfter(pt, "XML-INLINE");
					if (is != NULL)
					{
						is->TextIn = pt->TextAfter;
						pt->TextAfter = NULL;
					}
				}

				pt = pt->Next;
			}
		}

		if (TrivialInlineOrBlock != NULL)
		{
			if (Font != NULL)
			{
				AddTreePropertyValue(TrivialInlineOrBlock, "XML-FONT-FACE",
									 Font, ProcessSubFontTagAfterHit, false);
			}

			if (Size != NULL)
			{
				AddTreePropertyValue(TrivialInlineOrBlock, "XML-FONT-SIZE",
									 Size, ProcessSubFontTagAfterHit, false);
			}
			else
			{
				AddTreePropertyValue(TrivialInlineOrBlock, "XML-FONT-SIZE",
									 "10pt", ProcessSubFontTagAfterHit, false);
			}

			if (Color != NULL)
				AddTreePropertyValue(TrivialInlineOrBlock,
									 "XML-FONT-COLOR", Color,
									 ProcessSubFontTagAfterHit, false);

			if (Bold != false)
				AddTreePropertyValue(TrivialInlineOrBlock,
									 "XML-FONT-WEIGHT", "bold",
									 ProcessSubFontTagAfterHit, false);
			if (Underline != false)
				AddTreePropertyValue(TrivialInlineOrBlock,
									 "XML-FONT-TEXT-DECORATION",
									 "underline",
									 ProcessSubFontTagAfterHit, false);
			if (Italic != false)
				AddTreePropertyValue(TrivialInlineOrBlock,
									 "XML-FONT-STYLE", "italic",
									 ProcessSubFontTagAfterHit, false);
		}

		if (InParagraph != NULL)
		{
			if (LineHeight != NULL)
			{
				AddTreePropertyValue(InParagraph, "XML-LINE-HEIGHT",
									 LineHeight, ProcessSubFontTagAfterHit,
									 false);
			}
			else
			{
				AddTreePropertyValue(InParagraph, "XML-LINE-HEIGHT",
									 "12pt", ProcessSubFontTagAfterHit, false);
			}
		}

		/* for logging font structure */

		fprintf(mylog, "%*s/%s\n", depth * 4, "", pttmp->Type->Name);

		return pttmp;
	}
	else
		return NULL;
}

char ProcessTagHTML(tPTag pt)
{
	tPTag ptps, ptlms, ptlmssub, ptlmssubsub;
	tPProperty pp, pp2;
	int i;

	AddPropertyValue(pt, "XML-NAMESPACE-fox",
					 strdup("http://xml.apache.org/fop/extensions"));
	AddPropertyValue(pt, "XML-NAMESPACE-fo",
					 strdup("http://www.w3.org/1999/XSL/Format"));
	ptps = InsertAsSubTag(pt, "XML-PAGE-SEQUENCE");
	if (ptps != NULL)
	{

/*		AddPropertyValue(ptps,"XML-WORDWRAP","wrap");
*/

		AddPropertyValue(ptps, "XML-HYPHENATE", strdup("true"));

		pp = SearchProperty(pt, "LANG");

		if ((pp != NULL) && (pp->Value != NULL))
		{
			AddPropertyValue(ptps, "XML-PAGE-SEQUENCE-LANGUAGE",
							 strdup(pp->Value));

			pp2 = SearchProperty(ROOT, "XML-?-ENCODING");
			if ((pp2 != NULL) && (pp2->Value != NULL))
			{
				i = 0;
				while ((LangCodes[i][0] != NULL) &&
					   (strmulticmp(pp->Value, LangCodes[i][0]) != 0))
				{
					i++;
				}
				if (LangCodes[i][1] != NULL)
				{
					free(pp2->Value);
					pp2->Value = strdup(LangCodes[i][1]);
				}
			}

		}
		else
			AddPropertyValue(ptps, "XML-PAGE-SEQUENCE-LANGUAGE", strdup("en"));

		AddPropertyValue(ptps, "XML-PAGE-SEQUENCE-MASTER-REFERENCE",
						 strdup("first"));
	}

	ptlms = AddSubTagFirst(pt, "XML-LAYOUT-MASTER-SET");
	if (ptlms != NULL)
	{
		ptlms->TextIn = strdup("\r\n\t\t");
		ptlmssub = AddSubTagLast(ptlms, "XML-SIMPLE-PAGE-MASTER");
		if (ptlmssub != NULL)
		{
			ptlmssub->TextIn = strdup("\r\n\t\t\t");
			AddPropertyValue(ptlmssub, "XML-SIMPLE-PAGE-MASTER-NAME",
							 strdup("first"));
			AddPropertyValue(ptlmssub, "XML-PAGE-HEIGHT", strdup("29.7cm"));
			AddPropertyValue(ptlmssub, "XML-PAGE-WIDTH", strdup("21cm"));
			AddPropertyValue(ptlmssub, "XML-MARGIN-TOP", strdup("1.0cm"));
			AddPropertyValue(ptlmssub, "XML-MARGIN-BOTTOM", strdup("1.0cm"));
			AddPropertyValue(ptlmssub, "XML-MARGIN-LEFT", strdup("2.0cm"));
			AddPropertyValue(ptlmssub, "XML-MARGIN-RIGHT", strdup("2.0cm"));

			ptlmssubsub = AddSubTagLast(ptlmssub, "XML-REGION-BEFORE");
			if (ptlmssubsub != NULL)
			{
				AddPropertyValue(ptlmssubsub, "XML-REGION-EXTENT",
								 strdup("1.5cm"));
				ptlmssubsub->TextAfter = strdup("\r\n\t\t\t");
			}
			else
				fprintf(myout, "Internal Structure Error !!!\n");

			ptlmssubsub = AddSubTagLast(ptlmssub, "XML-REGION-BODY");
			if (ptlmssubsub != NULL)
			{
				AddPropertyValue(ptlmssubsub, "XML-MARGIN-TOP",
								 strdup("1.5cm"));
				AddPropertyValue(ptlmssubsub, "XML-MARGIN-BOTTOM",
								 strdup("1.5cm"));
				ptlmssubsub->TextAfter = strdup("\r\n\t\t\t");
			}
			else
				fprintf(myout, "Internal Structure Error !!!\n");

			ptlmssubsub = AddSubTagLast(ptlmssub, "XML-REGION-AFTER");
			if (ptlmssubsub != NULL)
			{
				AddPropertyValue(ptlmssubsub, "XML-REGION-EXTENT",
								 strdup("1.0cm"));
				ptlmssubsub->TextAfter = strdup("\r\n\t\t");
			}
			else
				fprintf(myout, "Internal Structure Error !!!\n");
			ptlmssub->TextAfter = strdup("\r\n\t");
		}
		else
			fprintf(myout, "Internal Structure Error !!!\n");
		ptlms->TextAfter = strdup("\r\n\t");
	}
	else
		fprintf(myout, "Internal Structure Error !!!\n");

	ptlms = AddSubTagFirst(pt, "!--");
	if (ptlms != NULL)
	{
		pt->TextIn = strdup("\r\n\t");
		AddProperty(ptlms, "--");
		AddPropertyValue(ptlms, "--Version", strdup(Version));
		AddPropertyValue(ptlms, "--Creator", strdup("html2fo"));
		ptlms->TextAfter = strdup("\r\n\t");
	}
	else
		fprintf(myout, "Internal Structure Error !!!\n");

	fprintf(myout, "analysing font structure...");
	fflush(myout);

	/* No property warnings while font processing */

	WarnProperties = false;

	SetupFont(0, pt, NULL, NULL, NULL, NULL, false, false, false, NULL, NULL);

	WarnProperties = true;

	fprintf(myout, " done\n");

	return 1;
}

char ProcessTagTitle(tPTag pTag)
{
	if (HTML_Title != NULL)
	{
		fprintf(myout,
				"I think you have two titles in your HTML Docoment...\nUsing the first one\n");
		free(pTag->TextIn);
	}
	else
	{
		HTML_Title = pTag->TextIn;
	}

	pTag->TextIn = NULL;

	return true;
}

char ProcessTagFooter(tPTag pTag)
{
	if (HTML_Footer != NULL)
	{
		fprintf(myout,
				"I think you have two footers in your HTML Docoment...\nUsing the first one\n");
		free(pTag->TextIn);
	}
	else
	{
		HTML_Footer = pTag->TextIn;
	}

	pTag->TextIn = NULL;

	return true;
}

char ProcessTagBodyBefore(tPTag pt)
{
	char s[10];

	sprintf(s, NumberFormat "cm", StdDocWidthcm);

	return (AddPropertyValue(pt, "XML-CONFIG-MAX-CONTENT-WIDTH", strdup(s))
			!= NULL);
}

char ProcessTagBody(tPTag pt)
{
	tPTag ptp, ptsc1, ptsc2;

	AddPropertyValue(pt, "BODY-XML-FLOW-NAME", strdup("xsl-region-body"));
	ptp = GetPrevTag(pt);
	if (ptp != NULL)
	{
		ptsc2 = AddTagAfter(ptp, "XML-STATIC-CONTENT");
		ptsc1 = AddTagAfter(ptp, "XML-STATIC-CONTENT");

	}
	else
	{
		ptsc2 = AddSubTagFirst(pt->Parent, "XML-STATIC-CONTENT");
		ptsc1 = AddSubTagFirst(pt->Parent, "XML-STATIC-CONTENT");
	}

	if (ptsc1 != NULL)
	{
		AddPropertyValue(ptsc1, "BODY-XML-FLOW-NAME",
						 strdup("xsl-region-before"));
		if (HTML_Title != NULL)
		{
			fprintf(myout, "Using Title \"%s\"\n", HTML_Title);
			ptsc1->TextIn = HTML_Title;
		}
		else
		{
			fprintf(myout, "Your HTML document has no Title - using default\n");
			ptsc1->TextIn =
				strdup
				("\r\n\r\n\t... the caption should be inserted here ...\r\n\r\n");
		}
		ptsc1->TextAfter = strdup("\r\n\r\n");
		AddSubTagFirstIncludeText(ptsc1, "P");
		ptsc1->TextIn = strdup("\r\n");
	}
	else
		fprintf(myout, "Internal Structure Error!!!\n");

	if (ptsc2 != NULL)
	{
		AddPropertyValue(ptsc2, "BODY-XML-FLOW-NAME",
						 strdup("xsl-region-after"));
		if (HTML_Footer != NULL)
		{
			fprintf(myout, "Using Footer \"%s\"\n", HTML_Footer);
			ptsc2->TextIn = HTML_Footer;
		}
		else
		{
			fprintf(myout,
					"Your HTML document has no Footer - using default\n");
			ptsc2->TextIn =
				strdup
				("\r\n\r\n\t... the footer should be inserted here ...\r\n\r\n");
		}
		ptsc2->TextAfter = strdup("\r\n\r\n");
		AddSubTagFirstIncludeText(ptsc2, "P");
		ptsc2->TextIn = strdup("\r\n");
	}
	else
		fprintf(myout, "Internal Structure Error!!!\n");

	ptp = AddSubTagLast(pt, "P");

	if (ptp != NULL)
	{
		AddPropertyValue(ptp, "XML-FONT-SIZE", strdup("1pt"));
		AddPropertyValue(ptp, "XML-LINE-HEIGHT", strdup("1pt"));
		AddPropertyValue(ptp, "XML-ID", strdup("LastPage"));
	}
	else
		fprintf(myout, "Internal Structure Error!!!\n");

	fprintf(myout, "setting up paragraphs...");
	fflush(myout);

	SetupParagraph(0, pt);

	fprintf(myout, " done.\n");

	return 1;
}

char ProcessTagListItem(tPTag pTag)
{
	tPTag ptla, ptbo, lablock;

	ptbo = InsertAsSubTag(pTag, "XML-LIST-ITEM-BODY");
	AddPropertyValue(ptbo, "XML-start-indent", strdup("body-start()"));
	ptla = AddSubTagFirst(pTag, "XML-LIST-ITEM-LABEL");
	SetupParagraph(0, ptbo);
	lablock = InsertAsSubTag(ptla, "P");
	lablock->TextIn = strdup("-");
	ptla->TextAfter = strdup("\n");
	return 1;
}

char ProcessTagA(tPTag pTag)
{
	AddPropertyValue(pTag, "XML-FONT-TEXT-DECORATION", strdup("underline"));
	AddPropertyValue(pTag, "XML-FONT-COLOR", strdup("#0000ff"));
	return 1;
}

char ProcessTagHeader(tPTag pTag)
{
	int v, count, v2, v3;
	char outc[8], c;
	char Bold = false;
	tPProperty pp;

	/* it is ok if the char after my TagName (e.g. H1, H2...) is not readable */
	count = sscanf(pTag->Type->Name, "%c%i%c", &c, &v, &c);
	if (count == 2)
	{
		switch (v)
		{
		case 1:
			Bold = true;
		case 2:
			v2 = 16;
			break;
		case 3:
			Bold = true;
		case 4:
			v2 = 14;
			break;
		case 5:
			Bold = true;
		case 6:
			v2 = 12;
			break;
		default:
			v2 = 10 + v * 2;
		}
		v3 = v2 * 2;
		sprintf(&outc[0], "%ipt", v2);
		pp = AddPropertyValue(pTag, "XML-FONT-SIZE", strdup(&outc[0]));
		fprintf(mylog, "Heading %i FontSize %s\n", v, pp->Value);

		sprintf(&outc[0], "%ipt", v3);
		pp = AddPropertyValue(pTag, "XML-LINE-HEIGHT", strdup(&outc[0]));
		fprintf(mylog, "           LineHeight %s\n", pp->Value);

		if (Bold != false)
		{
			AddPropertyValue(pTag, "XML-FONT-WEIGHT", strdup("bold"));
			fprintf(mylog, "           Bold\n");
		}

		return 1;
	}
	else
	{
		fprintf(mylog, "No FontSize for Heading!\n");
		return 0;
	}
	return 1;
}

char ProcessTagNullSubInfo(tPTag pTag)
{
	if (pTag->TextIn != NULL)
		free(pTag->TextIn);
	pTag->TextIn = NULL;
	FreeTag(&(pTag->SubTags));
	pTag->SubTags = NULL;
	pTag->LastSubTag = NULL;
	return true;
}

double GetMaxWidth(tPTag pt)
{
	tPProperty pp = NULL;
	double f = StdDocWidthcm;

	if (pt != NULL)
	{
		pp = GetConfig(pt, "XML-CONFIG-max-content-width");
		if (pp != NULL)
		{
			f = a2f(pp->Value, StdDocWidthcm);
		}
	}

	return f;
}

